# chess_ui.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Define the User Interface in detail.

A number of PanedWindows display scrollable grids which show lists of records
from the open database, and others display records selected from these grids
in more detail.
"""
# Split out of chess.py several years after it became clear ChessUI deserves a
# module of it's own.

import os
import tkinter
import tkinter.ttk
import tkinter.font
import tkinter.messagebox
import tkinter.filedialog
import queue

from solentware_grid.core.dataclient import DataSource

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from .gamerow import make_ChessDBrowGame
from .positionrow import make_ChessDBrowPosition
from .cqlrow import make_ChessDBrowCQL
from .repertoirerow import make_ChessDBrowRepertoire
from .gamelistgrid import (
    TagRosterGrid,
    GamePositionGames,
    PartialPositionGames,
    RepertoireGrid,
    RepertoirePositionGames,
)
from .game import Game
from .board import Board
from .cqlgrid import CQLGrid
from . import constants, fonts
from ..core.filespec import (
    GAMES_FILE_DEF,
    PARTIAL_FILE_DEF,
    REPERTOIRE_FILE_DEF,
    ANALYSIS_FILE_DEF,
    VARIATION_FIELD_DEF,
    SELECTION_FILE_DEF,
    RULE_FIELD_DEF,
    EVENT_FIELD_DEF,
    SITE_FIELD_DEF,
    DATE_FIELD_DEF,
    ROUND_FIELD_DEF,
    WHITE_FIELD_DEF,
    BLACK_FIELD_DEF,
    RESULT_FIELD_DEF,
    OPENING_FIELD_DEF,
    PARTIALPOSITION_NAME_FIELD_DEF,
)
from .displayitems import DisplayItems
from ..core.chessrecord import ChessDBrecordAnalysis
from .querygrid import QueryGrid
from .queryrow import make_ChessDBrowQuery
from .score import ScoreNoGameException


class ChessUIError(Exception):
    """Raise exception in chess_ui module."""


class ChessUI(ExceptionHandler):
    """Define widgets which form the User Interface."""

    allow_filter = {
        EVENT_FIELD_DEF,
        SITE_FIELD_DEF,
        DATE_FIELD_DEF,
        ROUND_FIELD_DEF,
        WHITE_FIELD_DEF,
        BLACK_FIELD_DEF,
        RESULT_FIELD_DEF,
    }

    def __init__(self, panel, statusbar=None, uci=None, toolbarframe=None):
        """Create the elements of the chess GUI.

        panel: the Panedwindow at root of the User Interface.
        statusbar: objects created by ChessUI can put text in the statusbar.
        uci: ChessUI objects communicate with chess engines via uci object.
        toolbarframe: selection rule control widgets are placed here.
        """
        self.top_pw = panel
        self.database = None
        self.partialpositionds = None
        self.fullpositionds = None
        self.engineanalysisds = None
        self.selectionds = None
        self.statusbar = statusbar
        self.uci = uci
        self.show_analysis = True
        self.visible_scrollbars = True

        # Create widgets for toolbarframe (to left of Statusbar set by Chess).
        if toolbarframe is not None:
            self.tb_entry = tkinter.ttk.Entry(toolbarframe, width=20)
            self.tb_entry.pack(side=tkinter.LEFT)
            self.tb_moveto = tkinter.ttk.Button(
                toolbarframe, text="Move to", underline=0
            )
            self.tb_moveto.pack(side=tkinter.LEFT)
            self.tb_filter = tkinter.ttk.Button(
                toolbarframe, text="Filter", underline=0
            )
            self.tb_filter.pack(side=tkinter.LEFT)
            self.set_toolbarframe_disabled()
        else:
            self.tb_entry = None
            self.tb_moveto = None
            self.tb_filter = None

        # All panes or just one with focus visible.
        self.single_view = False

        # So the parents can be given weight 1 when only one pane shown.
        self.pw_parent_map = {}

        # Pane weights when visible.
        # Top pane weight is 1 by definition: only item.
        # Consider removing top_pw settings as top_pw is not a subpane of a
        # PanedWindow.
        self.pw_weights = {}
        self.pw_current_weights = {}
        self.pw_weights[self.top_pw] = 1
        self.pw_current_weights[self.top_pw] = 0

        # So the payload widgets can be attached to their PanedWindows and
        # focus traversal can work when only one of the available payloads
        # is made visible.
        self.payload_parent_map = {}
        self.payload_available = {}

        self.error_file = None  # error log file name if available for use
        self._import_subprocess = None

        self.listfont = fonts.make_list_fonts(panel.winfo_toplevel())
        (
            self.tags_variations_comments_font,
            self.moves_played_in_game_font,
            self.boardfont,
            self.wildpieces,
        ) = fonts.make_chess_fonts(panel.winfo_toplevel())

        # The top pane has four panes arranged in columns.
        # Selection rules, database games list, displayed games, and analysis:
        # panes appear in this order from left to right when visible.
        self.selections_pw = tkinter.ttk.PanedWindow(
            self.top_pw, orient=tkinter.VERTICAL
        )
        self.list_games_pw = tkinter.ttk.PanedWindow(
            self.top_pw,
            # background='yellow',
            # opaqueresize=tkinter.FALSE,
            orient=tkinter.VERTICAL,
        )
        self.view_games_pw = tkinter.ttk.PanedWindow(
            self.top_pw,
            # background='yellow')
            orient=tkinter.VERTICAL,
        )
        self.analysis_pw = tkinter.ttk.PanedWindow(
            self.top_pw,
            # background='yellow',
            # opaqueresize=tkinter.FALSE,
            orient=tkinter.VERTICAL,
        )

        self.pw_parent_map[self.selections_pw] = self.top_pw
        self.pw_parent_map[self.list_games_pw] = self.top_pw
        self.pw_parent_map[self.view_games_pw] = self.top_pw
        self.pw_parent_map[self.analysis_pw] = self.top_pw

        # The analysis pane has two panes arranged in rows.
        self.repertoires_display_pw = tkinter.ttk.PanedWindow(
            self.analysis_pw, orient=tkinter.VERTICAL
        )
        self.partials_display_pw = tkinter.ttk.PanedWindow(
            self.analysis_pw, orient=tkinter.VERTICAL
        )
        self.pw_parent_map[self.repertoires_display_pw] = self.analysis_pw
        self.pw_parent_map[self.partials_display_pw] = self.analysis_pw

        # The lowest level PanedWindows.
        self.games_pw = tkinter.ttk.PanedWindow(
            self.list_games_pw, orient=tkinter.VERTICAL
        )
        self.position_games_pw = tkinter.ttk.PanedWindow(
            self.list_games_pw, orient=tkinter.VERTICAL
        )
        self.repertoires_pw = tkinter.ttk.PanedWindow(
            self.repertoires_display_pw, orient=tkinter.VERTICAL
        )
        self.position_repertoires_pw = tkinter.ttk.PanedWindow(
            self.repertoires_display_pw, orient=tkinter.VERTICAL
        )
        self.partials_pw = tkinter.ttk.PanedWindow(
            self.partials_display_pw, orient=tkinter.VERTICAL
        )
        self.position_partials_pw = tkinter.ttk.PanedWindow(
            self.partials_display_pw, orient=tkinter.VERTICAL
        )
        self.selection_rules_pw = tkinter.ttk.PanedWindow(
            self.selections_pw, orient=tkinter.VERTICAL
        )
        self.view_selection_rules_pw = tkinter.ttk.PanedWindow(
            self.selections_pw, orient=tkinter.VERTICAL
        )
        self.pw_parent_map[self.games_pw] = self.list_games_pw
        self.pw_parent_map[self.position_games_pw] = self.list_games_pw
        self.pw_parent_map[self.repertoires_pw] = self.repertoires_display_pw
        self.pw_parent_map[
            self.position_repertoires_pw
        ] = self.repertoires_display_pw
        self.pw_parent_map[self.partials_pw] = self.partials_display_pw
        self.pw_parent_map[
            self.position_partials_pw
        ] = self.partials_display_pw
        self.pw_parent_map[self.selection_rules_pw] = self.selections_pw
        self.pw_parent_map[self.view_selection_rules_pw] = self.selections_pw

        # These two, with self.view_games.pw bound earlier, contain the widgets
        # used to view the detail of games, repertoires, and partial positions.
        self.view_repertoires_pw = tkinter.ttk.PanedWindow(
            self.repertoires_display_pw, orient=tkinter.VERTICAL
        )
        self.view_partials_pw = tkinter.ttk.PanedWindow(
            self.partials_display_pw, orient=tkinter.VERTICAL
        )
        self.pw_parent_map[
            self.view_repertoires_pw
        ] = self.repertoires_display_pw
        self.pw_parent_map[self.view_partials_pw] = self.partials_display_pw

        # These display the lists of games, repertoires, partial positions, and
        # game selection rules, on the database.
        self.base_games = TagRosterGrid(self)
        self.base_repertoires = RepertoireGrid(self)
        self.base_partials = CQLGrid(self)
        self.base_selections = QueryGrid(self)

        # These display the lists of games, repertoires, and partial positions,
        # which match the current position of the active displayed item.
        self.game_games = GamePositionGames(self)
        self.repertoire_games = RepertoirePositionGames(self)
        self.partial_games = PartialPositionGames(self)

        def add_panedwindow(widget, weight):
            self.pw_current_weights[widget] = 0
            self.pw_parent_map[widget].add(
                widget, weight=self.pw_current_weights[widget]
            )
            self.pw_weights[widget] = weight

        # Selection columnn
        add_panedwindow(self.selections_pw, 1)

        # Database columnn
        add_panedwindow(self.list_games_pw, 2)

        # Games column (only item a container for managing subitems)
        add_panedwindow(self.view_games_pw, 4)

        # Analysis column
        add_panedwindow(self.analysis_pw, 2)

        # Selection column has two items, ratio 1:2
        add_panedwindow(self.selection_rules_pw, 1)
        add_panedwindow(self.view_selection_rules_pw, 2)

        # Database column has two items, ratio 2:1
        add_panedwindow(self.games_pw, 2)
        add_panedwindow(self.position_games_pw, 1)

        # Analysis column has two items, ratio 3:2
        add_panedwindow(self.repertoires_display_pw, 3)
        add_panedwindow(self.partials_display_pw, 2)

        # Upper item in analysis column has three items, ratio 1:2:1
        add_panedwindow(self.repertoires_pw, 1)
        add_panedwindow(self.view_repertoires_pw, 2)
        add_panedwindow(self.position_repertoires_pw, 1)

        # Lower item in analysis column has three items, ratio 1:1:1
        add_panedwindow(self.partials_pw, 1)
        add_panedwindow(self.view_partials_pw, 1)
        add_panedwindow(self.position_partials_pw, 1)

        # The managers for the four types of displayed item.
        self.game_items = DisplayItems()
        self.repertoire_items = DisplayItems()
        self.partial_items = DisplayItems()
        self.selection_items = DisplayItems()

        # The visibility of the repertoire and partial position lists when a
        # close action was done to a database: restore to this state on opening
        # a database.
        # Anything more complex will need something like DisplayItems.
        self.base_repertoires_displayed_at_database_close = False
        self.base_partials_displayed_at_database_close = False
        self.base_selections_displayed_at_database_close = False

        # Forward and backward traversal maps for next focus.
        self.traverse_forward = {
            self.base_games: self.game_games,
            self.game_games: self.game_items,
            self.game_items: self.base_repertoires,
            self.base_repertoires: self.repertoire_items,
            self.repertoire_items: self.repertoire_games,
            self.repertoire_games: self.base_partials,
            self.base_partials: self.partial_items,
            self.partial_items: self.partial_games,
            self.partial_games: self.base_selections,
            self.base_selections: self.selection_items,
            self.selection_items: self.base_games,
        }
        self.traverse_backward = {
            self.base_selections: self.partial_games,
            self.selection_items: self.base_selections,
            self.base_games: self.selection_items,
            self.game_games: self.base_games,
            self.game_items: self.game_games,
            self.base_repertoires: self.game_items,
            self.repertoire_items: self.base_repertoires,
            self.repertoire_games: self.repertoire_items,
            self.base_partials: self.repertoire_games,
            self.partial_items: self.base_partials,
            self.partial_games: self.partial_items,
        }

        def map_payload(payload, parent):
            self.payload_parent_map[payload] = parent
            self.payload_available[payload] = False

        # Map payload objects to their panedwindows, initially not available.
        # The widget is obtained from the payload's get_frame() method.
        map_payload(self.base_games, self.games_pw)
        map_payload(self.base_repertoires, self.repertoires_pw)
        map_payload(self.base_partials, self.partials_pw)
        map_payload(self.game_items, self.view_games_pw)
        map_payload(self.repertoire_items, self.view_repertoires_pw)
        map_payload(self.partial_items, self.view_partials_pw)
        map_payload(self.game_games, self.position_games_pw)
        map_payload(self.repertoire_games, self.position_repertoires_pw)
        map_payload(self.partial_games, self.position_partials_pw)
        map_payload(self.base_selections, self.selection_rules_pw)
        map_payload(self.selection_items, self.view_selection_rules_pw)

        # The game, repertoire, partial, and selection rule, widgets in their
        # own Toplevel.
        # Should these, added later, be helper class responsibility as well?
        # Similar construct in .uci.UCI too.
        self.games_and_repertoires_in_toplevels = set()
        self.partials_in_toplevels = set()
        self.selections_in_toplevels = set()

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under wine-1.8.1_1,1 on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # if self.uci.uci.uci_drivers_reply:
        #    self.uci.uci.set_analysis_queue(queue.Queue())
        #    self.process_uci_commands_from_engines_and_analysis_requests()
        # else:
        #    tkinter.messagebox.showinfo(
        #        'Chesstab Restriction',
        #        ' '.join(('Cannot create an interface for UCI engines:',
        #                  'this is expected if running under Wine.\n\n',
        #                  'You will not be allowed to start chess engines.')))
        self.uci.uci.set_analysis_queue(queue.Queue())
        self.process_uci_commands_from_engines_and_analysis_requests()

    def add_game_to_display(self, item):
        """Add game item to GUI."""
        self.game_items.add_item_to_display(item)
        if self.game_items.contains_one_item():

            # Call may not be necessary.
            # This one repeats call in gamedisplay.GameDisplay.
            # item.set_score_pointer_widget_navigation_bindings(True)

            self.set_game_position_data_source()
            self.calculate_payload_availability()
        self.configure_game_grid()
        self.set_game_change_notifications(item, callback=item.on_game_change)
        self.show_game_games()

    def add_partial_position_to_display(self, item):
        """Add partial position item to GUI."""
        self.partial_items.add_item_to_display(item)
        if self.partial_items.contains_one_item():
            self.set_partial_position_data_source()
            self.calculate_payload_availability()
            if self.database is not None:
                item.set_game_list()
        self.configure_partial_grid()
        self.set_partial_change_notifications(
            item, callback=item.on_partial_change
        )
        self.set_game_change_notifications(item, callback=item.on_game_change)
        self.show_partial_position_games()

    def add_repertoire_to_display(self, item):
        """Add repertoire item to GUI."""
        self.repertoire_items.add_item_to_display(item)
        if self.repertoire_items.contains_one_item():

            # Call may not be necessary.
            # item.set_score_pointer_widget_navigation_bindings(True)

            self.calculate_payload_availability()
            self.set_repertoire_data_source()
        self.configure_repertoire_grid()
        self.set_repertoire_change_notifications(
            item, callback=item.on_repertoire_change
        )
        self.set_game_change_notifications(item, callback=item.on_game_change)
        self.show_repertoire_games()

    def add_selection_rule_to_display(self, item):
        """Add selection rule item to GUI."""
        self.selection_items.add_item_to_display(item)
        if self.selection_items.contains_one_item():

            # Call may not be necessary.
            item.set_score_pointer_widget_navigation_bindings(True)

            self.calculate_payload_availability()
        self.configure_selection_grid()
        self.set_selection_rule_change_notifications(
            item, callback=item.on_selection_change
        )
        self.set_game_change_notifications(item, callback=item.on_game_change)
        self.show_selection_rule_games()

    def configure_game_grid(self):
        """Adjust game grid row sizes after navigate add or delete."""
        self.game_items.configure_items_grid(self.view_games_pw)

    def configure_partial_grid(self):
        """Adjust partial grid row sizes after navigate add or delete."""
        # If the constant is 2, like in configure_game_grid, sizing enters an
        # infinite loop if the first Partial Position is added when one or
        # more Repertoires are present.  The effect changes with constant, but
        # sizing finishes in all cases tried if it is Repertoire that is added
        # to one or more Partial Positions.  The work arounds are add another
        # display item of either type or drag resize the application.
        # There seems to be no problem if the constant is 1.
        # Baffling.
        # if len(self.partial_items.order) > 2:
        #    active_weight = len(self.partial_items.order) - 1
        # elif len(self.partialorder) == 0:
        #    # Closest to original code, but nothing gets done
        #    active_weight = 2
        # elif (self.partial_items.order[0].__class__ ==
        #      self.partial_items.order[-1].__class__):
        #    # So it looks like it's meant to do what it does.
        #    active_weight = 2
        # else:
        #    active_weight = 1
        # Back to original code when repertoire and partial position fully
        # separated.
        # The baffling problem may have been a consequence of references to
        # items displayed in list_games_pw (left-most) when deciding what to do
        # in analysis_pw (right-most).
        active_weight = max(2, len(self.partial_items.order) - 1)

        self.partial_items.configure_items_grid(
            self.view_partials_pw, active_weight=active_weight
        )

    def configure_repertoire_grid(self):
        """Adjust repertoire grid row sizes after navigate add or delete."""
        self.repertoire_items.configure_items_grid(self.view_repertoires_pw)

    def configure_selection_grid(self):
        """Adjust selection grid row sizes after navigate add or delete."""
        self.selection_items.configure_items_grid(self.view_selection_rules_pw)

    def delete_game_view(self, item):
        """Remove (game) panel from GUI on 'delete' event from panel."""
        views = self.game_items
        grid_item = views.delete_item_counters(item)
        if grid_item:
            self.set_properties_on_all_game_grids(grid_item)
        had_focus = item.has_focus()
        was_active = views.delete_item(item)

        self.configure_game_grid()
        self.view_games_pw.grid_rowconfigure(
            len(views.order), weight=0, uniform=""
        )

        # To refresh grid if current key is first after next database open.
        if self.database is None and self.game_games.partial not in (
            None,
            False,
        ):
            self.game_games.set_partial_key()

        if views.count_items_in_stack():
            if was_active:
                views.active_item.set_game_list()
            if had_focus:
                views.active_item.takefocus_widget.focus_set()
        else:
            if self.database:
                self.game_games.close_client_cursor()

                self.game_games.get_data_source().get_full_position_games(None)
                # Line above is reasonable when dealing with a built record set.
                # Line below seemed hackish at first.
                # However it is probably correct and the dubious code is likely
                # the long established use of None as the initial and default
                # value of attribute partial in instances of _DataAccess class.
                self.game_games.set_partial_key(key=False)

            self.calculate_payload_availability()
            self.configure_panedwindows()
            self.hide_game_games()
            if self.database is not None:
                if self.base_games and self.base_games.is_visible():
                    self.base_games.frame.focus_set()
                    return
                if self.base_partials and self.base_partials.is_visible():
                    self.base_partials.frame.focus_set()
                    return
            if self.partial_items.count_items_in_stack() != 0:

                # BUG start
                # partial position not resized
                self.partial_items.active_item.takefocus_widget.focus_set()
                # BUG end

    def delete_position_view(self, item):
        """Remove (partial) panel from GUI on 'delete' event from panel."""
        views = self.partial_items
        grid_item = views.delete_item_counters(item)
        if grid_item:
            if grid_item in self.base_partials.objects:
                self.base_partials.set_properties(grid_item)
        had_focus = item.has_focus()
        was_active = views.delete_item(item)

        self.configure_partial_grid()
        self.view_partials_pw.grid_rowconfigure(
            len(views.order), weight=0, uniform=""
        )

        # To refresh grid if current key is first after next database open.
        if self.database is None:
            if self.partial_games.partial not in (None, False):
                self.partial_games.set_partial_key()

        if views.count_items_in_stack():
            if was_active:
                self._set_find_partial_name_games(-1)
                widget = views.active_item.panel
                widget.after(
                    1,
                    func=self.try_command(
                        views.active_item.refresh_game_list, widget
                    ),
                )
                widget.after(
                    2, func=self.try_command(self._set_partial_name, widget)
                )
            if had_focus:
                views.active_item.takefocus_widget.focus_set()
        elif self.database is not None:
            self.hide_partial_position_games()

    def delete_repertoire_view(self, item):
        """Remove (game) panel from GUI on 'delete' event from panel."""
        views = self.repertoire_items
        grid_item = views.delete_item_counters(item)
        if grid_item:
            if grid_item in self.base_repertoires.objects:
                self.base_repertoires.set_properties(grid_item)
        had_focus = item.has_focus()
        was_active = views.delete_item(item)

        self.configure_repertoire_grid()
        self.view_repertoires_pw.grid_rowconfigure(
            len(views.order), weight=0, uniform=""
        )

        # To refresh grid if current key is first after next database open.
        if self.database is None:
            if self.repertoire_games.partial not in (None, False):
                self.repertoire_games.set_partial_key()

        if views.count_items_in_stack():
            if was_active:
                views.active_item.set_game_list()
            if had_focus:
                views.active_item.takefocus_widget.focus_set()
        elif self.database is not None:
            self.hide_repertoire_games()

    def delete_selection_rule_view(self, item):
        """Remove (selection rule) panel on 'delete' event from panel."""
        views = self.selection_items
        grid_item = views.delete_item_counters(item)
        if grid_item:
            if grid_item in self.base_selections.objects:
                self.base_selections.set_properties(grid_item)
        had_focus = item.has_focus()
        was_active = views.delete_item(item)

        self.configure_selection_grid()
        self.view_selection_rules_pw.grid_rowconfigure(
            len(views.order), weight=0, uniform=""
        )

        if views.count_items_in_stack():
            if was_active:
                views.active_item.refresh_game_list()
            if had_focus:
                views.active_item.takefocus_widget.focus_set()
        elif self.database is not None:
            self.hide_selection_rules()
            # Show all games in base_games.

    def _set_find_partial_name_games(self, index):
        """Set status text to partial position name being searched."""
        self.statusbar.set_status_text(
            self.partial_items.get_stack_item(index).get_text_for_statusbar()
        )

    def _set_partial_name(self):
        """Set status text to active partial position name."""
        self.statusbar.set_status_text(
            self.partial_items.active_item.get_text_for_statusbar()
        )

    def _set_selection_name(self):
        """Set status text to active selection rule name."""
        self.statusbar.set_status_text(
            self.selection_items.active_item.get_text_for_statusbar()
        )

    def get_active_game_move(self):
        """Return current move context of game being navigated."""
        try:
            return self.game_items.active_item.get_current_move_context()
        except IndexError:
            return None

    def get_active_repertoire_move(self):
        """Return current move context of repertoire being navigated."""
        try:
            return self.repertoire_items.active_item.get_current_move_context()
        except IndexError:
            return None

    def _hide_base_games(self):
        """Hide widget containing list of games in database."""
        if not self.base_games.is_visible():
            return
        self.base_games.forget_payload(
            self.payload_parent_map[self.base_games]
        )
        self.base_games.load_new_partial_key(False)
        self.base_games.set_data_source()

    def hide_game_grid(self):
        """Hide widget containing list of games in database."""
        if not self.base_games.is_visible():
            return
        if self.payload_available[self.base_repertoires]:
            self.base_repertoires_displayed_at_database_close = True
        if self.payload_available[self.base_partials]:
            self.base_partials_displayed_at_database_close = True
        if self.payload_available[self.base_selections]:
            self.base_selections_displayed_at_database_close = True
        self._hide_base_selection_rules()
        self._hide_base_partials()
        self._hide_base_repertoires()
        self._hide_base_games()
        self._set_position_analysis_data_source_all_items()
        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def hide_game_games(self):
        """Hide widgets containing selected games."""
        if self.game_items.count_items_in_stack():
            return
        self.game_games.clear_grid_keys()

        # Is next comment, kept from earlier version of this method called
        # hide_middle_column(), still relevant?

        ## Why different to hide_partial_position_games and
        ## hide_repertoire_games?
        ## Problem surfaces when games shown again.
        ## Only style of the three which works here.

        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def hide_partial_position_games(self):
        """Hide widgets containing selected partial positions and games."""
        if self.partial_items.count_items_in_stack():
            return

        # Why different to hide_game_games and hide_repertoire_games?
        # Problem surfaces when games shown again.
        # Only style of the three which works here.
        self.partial_games.forget_payload(
            self.payload_parent_map[self.partial_games]
        )
        if self.partials_display_pw.pane(self.position_partials_pw, "weight"):
            self.partial_games.load_new_partial_key(False)
            self.partial_games.set_data_source()

        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def _hide_base_partials(self):
        """Hide widget containing list of partial positions in database."""
        if not self.base_partials.is_visible():
            return
        self.base_partials.forget_payload(
            self.payload_parent_map[self.base_partials]
        )
        self.base_partials.load_new_partial_key(False)
        self.base_partials.set_data_source()

    def hide_partial_position_grid(self):
        """Hide widget containing list of partial positions in database."""
        if not self.base_partials.is_visible():
            return
        self._hide_base_partials()
        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def hide_repertoire_games(self):
        """Hide widgets containing selected repertoires and games."""
        if self.repertoire_items.count_items_in_stack():
            return

        # Why different to hide_partial_position_games and hide_game_games?
        # Problem surfaces when games shown again.
        # Only style of the three which works here.
        if self.repertoires_display_pw.pane(
            self.position_repertoires_pw, "weight"
        ):
            self.repertoire_games.load_new_partial_key(False)
            self.repertoire_games.set_data_source()

        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def _hide_base_repertoires(self):
        """Hide widget containing list of repertoire games in database."""
        if not self.base_repertoires.is_visible():
            return
        self.base_repertoires.forget_payload(
            self.payload_parent_map[self.base_repertoires]
        )
        self.base_repertoires.load_new_partial_key(False)
        self.base_repertoires.set_data_source()

    def hide_repertoire_grid(self):
        """Hide widget containing list of repertoire games in database."""
        if not self.base_repertoires.is_visible():
            return
        self._hide_base_repertoires()
        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def _hide_base_selection_rules(self):
        """Hide widget containing list of selection rules in database."""
        if not self.base_selections.is_visible():
            return
        self.base_selections.forget_payload(
            self.payload_parent_map[self.base_selections]
        )
        self.base_selections.load_new_partial_key(False)
        self.base_selections.set_data_source()

    def hide_selection_rules_grid(self):
        """Hide widget containing list of selection rules in database."""
        if not self.base_selections.is_visible():
            return
        self._hide_base_selection_rules()
        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def hide_selection_rules(self):
        """Hide widgets containing selected selection rules."""
        if self.selection_items.count_items_in_stack():
            return
        self.calculate_payload_availability()
        self.configure_panedwindows()
        self.reset_focus_on_hide_widget()

    def reset_focus_on_hide_widget(self):
        """Switch focus when widget with focus is hidden."""
        # BUG start
        # There was a comment here on a problem whose detail was not recorded.
        # But if sashes have been dragged such that the widget being hidden
        # is not visible when being hiddden, the list of games on close database
        # for example, then w is None in the w.winfo_pathname(...) call.
        #
        # hidden is the wrong argument.  Should be the widgets to take focus in
        # preference order and if that does not give an answer go through the
        # list below.
        if self.base_games.is_visible():
            self.base_games.frame.focus_set()
        elif self.base_partials.is_visible():
            self.base_partials.frame.focus_set()
        elif self.base_repertoires.is_visible():
            self.base_repertoires.frame.focus_set()
        elif self.game_items.count_items_in_stack():
            self.game_items.active_item.takefocus_widget.focus_set()
        elif self.repertoire_items.count_items_in_stack():
            self.repertoire_items.active_item.takefocus_widget.focus_set()
        elif self.partial_items.count_items_in_stack():
            self.partial_items.active_item.takefocus_widget.focus_set()
        else:
            self.top_pw.focus_set()
        # BUG end

    def set_board_colours(self, colourscheme):
        """Set board colours to match colourscheme."""
        sbg = False
        if colourscheme.game.l_color != Game.l_color:
            Game.l_color = colourscheme.game.l_color
            sbg = True
        if colourscheme.game.m_color != Game.m_color:
            Game.m_color = colourscheme.game.m_color
            sbg = True
        if colourscheme.game.am_color != Game.am_color:
            Game.am_color = colourscheme.game.am_color
            sbg = True
        if colourscheme.game.v_color != Game.v_color:
            Game.v_color = colourscheme.game.v_color
            sbg = True
        bbg = False
        if colourscheme.game.board.litecolor != Board.litecolor:
            Board.litecolor = colourscheme.game.board.litecolor
            bbg = True
        if colourscheme.game.board.darkcolor != Board.darkcolor:
            Board.darkcolor = colourscheme.game.board.darkcolor
            bbg = True
        bfg = False
        if colourscheme.game.board.whitecolor != Board.whitecolor:
            Board.whitecolor = colourscheme.game.board.whitecolor
            bfg = True
        if colourscheme.game.board.blackcolor != Board.blackcolor:
            Board.blackcolor = colourscheme.game.board.blackcolor
            bfg = True
        self._set_board_colours(sbg, bbg, bfg)

    def set_board_colours_from_options(self, defaults):
        """Set colour scheme board on opening database."""
        if not defaults:
            return
        sbg = False
        bbg = False
        bfg = False
        if defaults[constants.LINE_COLOR_NAME]:
            Game.l_color = defaults[constants.LINE_COLOR_NAME]
            sbg = True
        if defaults[constants.MOVE_COLOR_NAME]:
            Game.m_color = defaults[constants.MOVE_COLOR_NAME]
            sbg = True
        if defaults[constants.ALTERNATIVE_MOVE_COLOR_NAME]:
            Game.am_color = defaults[constants.ALTERNATIVE_MOVE_COLOR_NAME]
            sbg = True
        if defaults[constants.VARIATION_COLOR_NAME]:
            Game.v_color = defaults[constants.VARIATION_COLOR_NAME]
            sbg = True
        if defaults[constants.LITECOLOR_NAME]:
            Board.litecolor = defaults[constants.LITECOLOR_NAME]
            bbg = True
        if defaults[constants.DARKCOLOR_NAME]:
            Board.darkcolor = defaults[constants.DARKCOLOR_NAME]
            bbg = True
        if defaults[constants.WHITECOLOR_NAME]:
            Board.whitecolor = defaults[constants.WHITECOLOR_NAME]
            bfg = True
        if defaults[constants.BLACKCOLOR_NAME]:
            Board.blackcolor = defaults[constants.BLACKCOLOR_NAME]
            bfg = True
        self._set_board_colours(sbg, bbg, bfg)
        if defaults[constants.MOVES_PLAYED_IN_GAME_FONT]:
            fonts.copy_font_attributes(
                defaults[constants.MOVES_PLAYED_IN_GAME_FONT],
                tkinter.font.nametofont(constants.MOVES_PLAYED_IN_GAME_FONT),
            )
        if defaults[constants.TAGS_VARIATIONS_COMMENTS_FONT]:
            fonts.copy_font_attributes(
                defaults[constants.TAGS_VARIATIONS_COMMENTS_FONT],
                tkinter.font.nametofont(constants.MOVES_PLAYED_IN_GAME_FONT),
            )
        if defaults[constants.PIECES_ON_BOARD_FONT]:
            fonts.copy_board_font_attributes(
                defaults[constants.PIECES_ON_BOARD_FONT],
                tkinter.font.nametofont(constants.PIECES_ON_BOARD_FONT),
            )

    def set_board_fonts(self, colourscheme):
        """Set board fonts to match colourscheme."""
        exceptions = []
        capobftf = colourscheme.apply_pieces_on_board_font_to_font
        for games in (
            self.game_items.order,
            self.partial_items.order,
            self.repertoire_items.order,
            self.games_and_repertoires_in_toplevels,
            self.partials_in_toplevels,
        ):
            for item in games:
                try:
                    capobftf(item.board.font)
                except tkinter.TclError:
                    exceptions.append((item, games))
        for item, games in exceptions:
            games.remove(item)

    @staticmethod
    def set_data_change_notifications(widget, notifiers, callback=None):
        """Enable or disable database update notifications for widget.

        widget - object to be notified
        notifiers - sequence of objects which do the notifying
        callback - function, usually a method of widget, used to do notifying

        if callback is None the specified notifications are cancelled
        """
        if callback is None:
            for grid in notifiers:
                widget.register_out(grid.get_data_source())
        else:
            for grid in notifiers:
                widget.register_in(grid.get_data_source(), callback)

    def set_game_change_notifications(self, widget, callback=None):
        """Set or unset game update notifications to widget."""
        self.set_data_change_notifications(
            widget,
            (self.base_games, self.game_games, self.partial_games),
            callback,
        )

    def set_partial_change_notifications(self, widget, callback=None):
        """Set or unset partial position update notifications to widget."""
        self.set_data_change_notifications(
            widget, (self.base_partials,), callback
        )

    def set_repertoire_change_notifications(self, widget, callback=None):
        """Set or unset partial position update notifications to widget."""
        self.set_data_change_notifications(
            widget, (self.base_repertoires,), callback
        )

    def set_selection_rule_change_notifications(self, widget, callback=None):
        """Set or unset selection rule update notifications to widget."""
        self.set_data_change_notifications(
            widget, (self.base_selections,), callback
        )

    def set_open_database_and_engine_classes(
        self,
        database=None,
        fullpositionclass=None,
        partialpositionclass=None,
        engineanalysisclass=None,
        selectionclass=None,
    ):
        """Set current open database and engine specific dataset classes."""
        self.database = database
        if database is None:
            self.partialpositionds = None
            self.fullpositionds = None
            self.engineanalysisds = None
            self.selectionds = None
            self.uci.uci.set_position_analysis_data_source()
            self.uci.set_open_database_and_engine_classes()
        else:
            self.partialpositionds = partialpositionclass
            self.fullpositionds = fullpositionclass
            self.engineanalysisds = engineanalysisclass
            self.selectionds = selectionclass
            self.uci.uci.set_position_analysis_data_source(
                datasource=self.make_position_analysis_data_source()
            )
            self.uci.set_open_database_and_engine_classes(database=database)

    def set_focus_gamepanel_item(self, event=None):
        """Give game at top of stack the focus."""
        self.set_toolbarframe_disabled()
        self.game_items.set_focus()

    def set_focus_game_grid(self, event=None):
        """Give widget displaying list of games on database the focus."""
        if self.database is not None:
            if self.base_games.datasource.dbname in self.allow_filter:
                self.set_toolbarframe_normal(
                    self.move_to_game, self.filter_game
                )
            else:
                self.set_toolbarframe_disabled()
            self.base_games.set_focus()
            self.base_games.set_selection_text()

    def set_focus_partial_game_grid(self, event=None):
        """Give widget displaying list of games for partial position focus."""
        if self.database is not None:
            self.set_toolbarframe_disabled()
            self.partial_games.set_focus()
            self.partial_games.set_selection_text()

    def set_focus_partial_grid(self, event=None):
        """Give widget displaying list of partial positions the focus."""
        if self.database is not None:
            self.set_toolbarframe_normal(
                self.move_to_partial, self.filter_partial
            )
            self.base_partials.set_focus()
            self.base_partials.set_selection_text()

    def set_focus_partialpanel_item(self, event=None):
        """Give partial position at top of stack the focus."""
        self.set_toolbarframe_disabled()
        self.partial_items.set_focus()

    def set_focus_position_grid(self, event=None):
        """Give widget displaying list of games matching position the focus."""
        if self.database is not None:
            self.set_toolbarframe_disabled()
            self.game_games.set_focus()
            self.game_games.set_selection_text()

    def set_focus_repertoire_game_grid(self, event=None):
        """Give widget displaying list of games for repertoire focus."""
        if self.database is not None:
            self.set_toolbarframe_disabled()
            self.repertoire_games.set_focus()
            self.repertoire_games.set_selection_text()

    def set_focus_repertoire_grid(self, event=None):
        """Give focus to widget displaying list of database repertoire games."""
        if self.database is not None:
            self.set_toolbarframe_normal(
                self.move_to_repertoire, self.filter_repertoire
            )
            self.base_repertoires.set_focus()
            self.base_repertoires.set_selection_text()

    def set_focus_repertoirepanel_item(self, event=None):
        """Give repertoire game at top of stack the focus."""
        self.set_toolbarframe_disabled()
        self.repertoire_items.set_focus()

    def set_focus_selection_rule_grid(self, event=None):
        """Give widget displaying list of selection rules the focus."""
        if self.database is not None:
            self.set_toolbarframe_normal(
                self.move_to_selection, self.filter_selection
            )
            self.base_selections.set_focus()
            self.base_selections.set_selection_text()

    def set_focus_selectionpanel_item(self, event=None):
        """Give selection rule at top of stack the focus."""
        self.set_toolbarframe_disabled()
        self.selection_items.set_focus()

    def set_game_position_data_source(self):
        """Attach database to widget that displays games for position."""
        if self.database is not None:
            sqds = self.fullpositionds(
                self.database,
                GAMES_FILE_DEF,
                GAMES_FILE_DEF,
                make_ChessDBrowPosition(self),
            )
        else:
            sqds = None
        self.game_games.set_data_source(sqds)

    def set_partial_position_data_source(self):
        """Attach database to widget displaying games for partial position."""
        if self.database is not None:
            sqds = self.partialpositionds(
                self.database,
                GAMES_FILE_DEF,
                GAMES_FILE_DEF,
                make_ChessDBrowGame(self),
            )
        else:
            sqds = None

        # The expected effect, using correct header, does not happen.
        # self.partial_games.set_data_header(header=ChessDBrowGame)
        # In some runs always does right thing but in others gives empty grid.
        # Seems to be wrong always if partial item is asked for first.
        # self.partial_games.make_header(ChessDBrowGame.header_specification)
        self.partial_games.set_data_source(sqds)

    def set_repertoire_data_source(self):
        """Attach database to widget displaying games for repertoire."""
        if self.database is not None:
            sqds = self.fullpositionds(
                self.database,
                GAMES_FILE_DEF,
                GAMES_FILE_DEF,
                make_ChessDBrowPosition(self),
            )
        else:
            sqds = None

        # The expected effect, using correct header, does not happen.
        # self.partial_games.set_data_header(header=ChessDBrowPosition)
        # In some runs always does right thing but in others gives empty grid.
        # Seems to be correct always if repertoire item is asked for first,
        # but only if database is open when first is requested.
        # self.partial_games.make_header(
        #    ChessDBrowPosition.header_specification)
        self.repertoire_games.set_data_source(sqds)

    def _show_base_games(self):
        """Show widget containing list of games in database."""
        if self.list_games_pw.pane(self.games_pw, "weight"):

            # Size is assumed to be (1,1) as created at first call and grid
            # will be filled after <Configure> event to size the widget.
            # Subsequent calls are not associated with a <Configure> event so
            # fill the widget by after_idle(...).
            if self.base_games.get_frame().winfo_width() != 1:
                self.games_pw.after_idle(self.base_games.load_new_index)

    @staticmethod
    def _show_active_item_position_games(active):
        """Show list for current position in active item."""
        if not active:
            return

        # Test is on active.score because active.analysis does not exist if
        # active is from self.partial_items.
        if active.takefocus_widget is not active.score:
            active = active.analysis

        active.itemgrid.set_partial_key()
        active.set_game_list()

    def show_game_grid(self, database):
        """Show widget containing list of games in database."""
        if self.game_items.count_items_in_stack():
            self.set_game_position_data_source()
            self.game_items.active_item.set_game_list()
            self.game_items.set_insert_or_delete_on_all_items()
        if self.partial_items.count_items_in_stack():
            self.set_partial_position_data_source()
            self.partial_items.active_item.refresh_game_list()
            self.partial_items.set_insert_or_delete_on_all_items()
        if self.repertoire_items.count_items_in_stack():
            self.set_repertoire_data_source()
            self.repertoire_items.active_item.set_game_list()
            self.repertoire_items.set_insert_or_delete_on_all_items()
        if self.selection_items.count_items_in_stack():
            for i in self.selection_items.order:
                if i.query_statement.dbset is None:
                    i.query_statement.dbset = GAMES_FILE_DEF
            # self.set_repertoire_data_source()
            # self.selection_items.active_item.set_game_list()
            self.selection_items.set_insert_or_delete_on_all_items()
        self._set_position_analysis_data_source_all_items()

        ## Hack to cope with GUI use while Import in progress.
        # if self.base_games.get_data_source().get_database() is not None:
        #    self.base_games.clear_grid_keys()

        if self.base_repertoires_displayed_at_database_close:
            self._create_repertoire_datasource(database)
        if self.base_partials_displayed_at_database_close:
            self._create_partial_position_datasource(database)
        if self.base_selections_displayed_at_database_close:
            self._create_selection_rules_datasource(database)
        self.base_repertoires_displayed_at_database_close = False
        self.base_partials_displayed_at_database_close = False
        self.base_selections_displayed_at_database_close = False
        self.calculate_payload_availability()
        self.configure_panedwindows()
        self._show_base_games()
        self._show_base_partials()
        self._show_base_repertoires()
        self._show_base_selection_rules()

        # All the active items except selection_items, which is excluded so all
        # games in database are listed when database is opened.
        self._show_active_item_position_games(self.game_items.active_item)
        self._show_active_item_position_games(
            self.repertoire_items.active_item
        )
        self._show_active_item_position_games(self.partial_items.active_item)

        self.base_games.set_focus()

        # Refresh.
        self.set_game_change_notifications(
            self.base_games, callback=self.base_games.on_game_change
        )

    def show_game_games(self):
        """List games containing position in active game."""
        if not self.game_items.count_items_in_stack():
            return

        ## Hack to cope with GUI use while Import in progress.
        ##if self.partial_games.get_data_source() is not None:
        ##    if self.partial_games.get_data_source(
        ##        ).get_database() is not None:
        ##        self.partial_games.clear_grid_keys()
        # if self.game_games.get_data_source() is not None:
        #    if self.game_games.get_data_source(
        #        ).get_database() is not None:
        #        self.game_games.clear_grid_keys()

        if self.top_pw.pane(self.list_games_pw, "weight"):
            self.set_game_change_notifications(
                self.base_games, callback=self.base_games.on_game_change
            )
        self.configure_panedwindows()

    def show_partial_position_games(self):
        """List games containing partial position in active partial position."""
        if not self.partial_items.count_items_in_stack():
            if not self.top_pw.pane(self.analysis_pw, "weight"):
                return

        ## Hack to cope with GUI use while Import in progress.
        # if self.partial_games.get_data_source() is not None:
        #    if self.partial_games.get_data_source(
        #        ).get_database() is not None:
        #        self.partial_games.clear_grid_keys()

        if not self.partials_display_pw.pane(self.view_partials_pw, "weight"):
            self.set_game_change_notifications(
                self.base_games, callback=self.base_games.on_game_change
            )
        self.configure_panedwindows()

    def show_repertoire_games(self):
        """List games containing position in active repertoire."""
        if not self.repertoire_items.count_items_in_stack():
            if not self.top_pw.pane(self.analysis_pw, "weight"):
                return

        ## Hack to cope with GUI use while Import in progress.
        # if self.repertoire_games.get_data_source() is not None:
        #    if self.repertoire_games.get_data_source(
        #        ).get_database() is not None:
        #        self.repertoire_games.clear_grid_keys()

        if not self.repertoires_display_pw.pane(
            self.view_repertoires_pw, "weight"
        ):
            self.set_game_change_notifications(
                self.base_games, callback=self.base_games.on_game_change
            )
        self.configure_panedwindows()

    # Not sure how correct this is yet: in pane hierarchy terms.
    def show_selection_rule_games(self):
        """List games containing position in active repertoire."""
        if not self.selection_items.count_items_in_stack():
            if not self.top_pw.pane(self.selections_pw, "weight"):
                return

        ## Hack to cope with GUI use while Import in progress.
        # if self.repertoire_games.get_data_source() is not None:
        #    if self.repertoire_games.get_data_source(
        #        ).get_database() is not None:
        #        self.repertoire_games.clear_grid_keys()

        if not self.selections_pw.pane(self.selection_rules_pw, "weight"):
            self.set_game_change_notifications(
                self.base_games, callback=self.base_games.on_game_change
            )
        self.configure_panedwindows()

    def _create_partial_position_datasource(self, database):
        """Create a new DataSource for list of partial positions."""
        # Probably wrong because datasource = F(active item), but this is a
        # constant at present.
        self.base_partials.set_data_source(
            DataSource(
                database,
                PARTIAL_FILE_DEF,
                PARTIALPOSITION_NAME_FIELD_DEF,
                make_ChessDBrowCQL(self),
            ),
            self.base_partials.on_data_change,
        )
        if self.partial_items.any_items_displayed_of_type():
            self.set_partial_position_data_source()
            self.partial_items.set_insert_or_delete_on_all_items()

    def _show_base_partials(self):
        """Show widget containing list of partial positions in database."""
        if self.partials_display_pw.pane(self.partials_pw, "weight"):

            # Size is assumed to be (1,1) as created at first call and grid
            # will be filled after <Configure> event to size the widget.
            # Subsequent calls are not associated with a <Configure> event so
            # fill the widget by after_idle(...).
            if self.base_partials.get_frame().winfo_width() != 1:
                self.base_partials.set_partial_key()
                self.partials_pw.after_idle(self.base_partials.load_new_index)

    def show_partial_position_grid(self, database):
        """Show widget containing list of partial positions in database."""
        self._create_partial_position_datasource(database)

        ## Hack to cope with GUI use while Import in progress.
        # if self.base_partials.get_data_source().get_database() is not None:
        #    self.base_partials.clear_grid_keys()

        self.calculate_payload_availability()
        self.configure_panedwindows()
        self._show_base_partials()
        self.base_partials.set_focus()

    def _create_repertoire_datasource(self, database):
        """Create a new DataSource for list of repertoires."""
        # Probably wrong because datasource = F(active item), but this is a
        # constant at present.
        self.base_repertoires.set_data_source(
            DataSource(
                database,
                REPERTOIRE_FILE_DEF,
                OPENING_FIELD_DEF,
                make_ChessDBrowRepertoire(self),
            ),
            self.base_repertoires.on_data_change,
        )
        if self.repertoire_items.any_items_displayed_of_type():
            self.set_repertoire_data_source()
            self.repertoire_items.set_insert_or_delete_on_all_items()

    def _show_base_repertoires(self):
        """Show widget containing list of repertoire games in database."""
        if self.repertoires_display_pw.pane(self.repertoires_pw, "weight"):

            # Size is assumed to be (1,1) as created at first call and grid
            # will be filled after <Configure> event to size the widget.
            # Subsequent calls are not associated with a <Configure> event so
            # fill the widget by after_idle(...).
            if self.base_repertoires.get_frame().winfo_width() != 1:
                self.base_repertoires.set_partial_key()
                self.repertoires_pw.after_idle(
                    self.base_repertoires.load_new_index
                )

    def show_repertoire_grid(self, database):
        """Show widget containing list of repertoire games in database."""
        self._create_repertoire_datasource(database)

        ## Hack to cope with GUI use while Import in progress.
        # if self.base_repertoires.get_data_source().get_database() is not None:
        #    self.base_repertoires.clear_grid_keys()

        self.calculate_payload_availability()
        self.configure_panedwindows()
        self._show_base_repertoires()
        self.base_repertoires.set_focus()

    def set_import_subprocess(self, subprocess_id=None):
        """Set the import subprocess object if not already active."""
        if self.is_import_subprocess_active():
            raise ChessUIError("Attempt to set import subprocess while active")
        self._import_subprocess = subprocess_id

    def get_import_subprocess(self):
        """Return the import subprocess identity."""
        return self._import_subprocess

    def get_import_subprocess_poll(self):
        """Poll the import subprocess and return the response."""
        return self._import_subprocess.poll()

    def is_import_subprocess_active(self):
        """Return True if the import subprocess object is active."""
        if self._import_subprocess is None:
            return False
        return self._import_subprocess.poll() is None

    def get_toplevel(self):
        """Return the toplevel widget."""
        return self.top_pw.winfo_toplevel()

    def process_uci_commands_from_engines_and_analysis_requests(self):
        """Periodically process UCI actions required and response queues."""
        self.uci.uci.get_engine_responses()
        self.uci.uci.get_analysis_requests()
        self.refresh_analysis_widgets()
        self.top_pw.after(
            1000, self.process_uci_commands_from_engines_and_analysis_requests
        )

    @staticmethod
    def _configure_font(target, source):
        """Set target font properties from source font."""
        for property_ in "family", "weight", "slant", "size":
            target[property_] = source[property_]

    def _set_board_colours(self, sbg, bbg, bfg):
        """Set colours and fonts used to display games.

        sbg == True - set game score colours
        bbg == True - set board square colours
        bfg == True - set board piece colours

        """
        exceptions = []
        for games in (
            self.game_items.order,
            self.partial_items.order,
            self.repertoire_items.order,
            self.games_and_repertoires_in_toplevels,
            self.partials_in_toplevels,
        ):
            for item in games:
                try:
                    item.set_colours(sbg, bbg, bfg)
                except tkinter.TclError:
                    exceptions.append((item, games))
        for item, games in exceptions:
            games.remove(item)

    def get_export_filename(self, datatype, pgn=True):
        """Return filename to contain export of datatype or None."""
        title = " ".join(("Export", datatype))
        if self.is_import_subprocess_active():
            tkinter.messagebox.showinfo(
                parent=self.get_toplevel(),
                title=title,
                message="An import of data is in progress",
            )
            return None
        if not self.database:
            tkinter.messagebox.showinfo(
                parent=self.get_toplevel(),
                title=title,
                message="".join(
                    ("Open the database from which the export is to be done.",)
                ),
            )
            return None
        return self._get_export_filename(datatype, pgn, title)

    def get_export_filename_for_single_item(self, datatype, pgn=True):
        """Return filename to contain export of datatype or None."""
        return self._get_export_filename(
            datatype, pgn, " ".join(("Export", datatype))
        )

    def get_export_folder(self):
        """Return folder name to contain export of database or None."""
        title = "Export database as Text"
        if self.is_import_subprocess_active():
            tkinter.messagebox.showinfo(
                parent=self.get_toplevel(),
                title=title,
                message="An import of data is in progress",
            )
            return None
        if not self.database:
            tkinter.messagebox.showinfo(
                parent=self.get_toplevel(),
                title=title,
                message="".join(
                    ("Open the database from which the export is to be done.",)
                ),
            )
            return None
        filename = tkinter.filedialog.askdirectory(
            parent=self.get_toplevel(), title=title, initialdir="~"
        )
        if not filename:
            tkinter.messagebox.showwarning(
                parent=self.get_toplevel(),
                title=title,
                message="Database not exported",
            )
            return None
        bfn = os.path.basename(filename)
        fns = (
            os.path.join(filename, "".join((bfn, "_games", ".txt"))),
            os.path.join(filename, "".join((bfn, "_repertoires", ".txt"))),
            os.path.join(filename, "".join((bfn, "_partials", ".txt"))),
        )
        if not tkinter.messagebox.askokcancel(
            parent=self.get_toplevel(),
            title=title,
            message="".join(
                (
                    "The database will be exported to files:\n\n",
                    "\n\n".join((os.path.basename(f) for f in fns)),
                    "\n\nprovided none of these already exist.\n",
                )
            ),
        ):
            return None
        for filename in fns:
            if os.path.exists(filename):
                tkinter.messagebox.showinfo(
                    parent=self.get_toplevel(),
                    title=title,
                    message="\n".join(
                        (
                            "Cannot export because file\n",
                            filename,
                            "\nalready exists.\n",
                        )
                    ),
                )
                return None
        return fns

    def _get_export_filename(self, datatype, pgn, title):
        """Return filename to contain export of datatype or None."""
        if pgn:
            extn = "pgn"
        else:
            extn = "txt"
        filename = tkinter.filedialog.asksaveasfilename(
            parent=self.get_toplevel(),
            title=title,
            defaultextension="".join((".", extn)),
            filetypes=((datatype, ".".join(("*", extn))),),
        )
        if not filename:
            tkinter.messagebox.showwarning(
                parent=self.get_toplevel(),
                title=title,
                message=" ".join((datatype, "file not saved")),
            )
            return None
        return filename

    def refresh_analysis_widgets(self):
        """Refresh game widgets with updated chess engine analysis."""
        exceptions = []
        for games in (
            self.game_items.order,
            self.repertoire_items.order,
            self.games_and_repertoires_in_toplevels,
        ):
            for item in games:
                try:
                    if item.current is None:
                        try:
                            position = item.fen_tag_tuple_square_piece_map()
                        except ScoreNoGameException:
                            continue
                    else:
                        position = item.tagpositionmap[item.current]
                    item.refresh_analysis_widget(item.get_analysis(*position))
                except tkinter.TclError:
                    exceptions.append((item, games))
        for item, games in exceptions:
            games.remove(item)

    @staticmethod
    def _give_focus(current, traverse):
        """Give focus to adjacent widget type in traversal order."""
        give = current
        while True:
            give = traverse[give]
            if give == current:
                break
            if give.is_visible():
                give.set_focus()
                break

    def give_focus_backward(self, current):
        """Give focus to previous widget type in traversal order."""
        self._give_focus(current, self.traverse_backward)

    def give_focus_forward(self, current):
        """Give focus to next widget type in traversal order."""
        self._give_focus(current, self.traverse_forward)

    def show_all_panedwindows(self):
        """Show all panes with weights prior to toggle to single pane view."""
        if not self.single_view:
            return
        for widget, parent in self.pw_parent_map.items():
            parent.pane(widget, weight=self.pw_current_weights[widget])
        self.single_view = False
        self.configure_game_grid()
        self.configure_partial_grid()
        self.configure_repertoire_grid()
        self.configure_selection_grid()

    def show_just_panedwindow_with_focus(self, gainfocus):
        """Show pane containing widget with focus and hide the other panes."""
        ppmi = {v: k for k, v in self.payload_parent_map.items()}
        widget = gainfocus
        while True:
            if widget in ppmi:
                if widget in self.pw_parent_map:
                    self.single_view = True
                    self.single_view = self.show_payload_panedwindows(
                        ppmi[widget]
                    )
                break
            if widget is self.top_pw:
                break
            widget = self.top_pw.nametowidget(widget.winfo_parent())
        if self.single_view:
            self.configure_game_grid()
            self.configure_partial_grid()
            self.configure_repertoire_grid()

    def show_payload_panedwindows(self, payload):
        """Show just payload's ancestor's panes if payload available."""
        if not self.single_view:
            return False
        if not self.payload_available[payload]:
            return False
        for ppw, parent in self.pw_parent_map.items():
            parent.pane(ppw, weight=0)
        ppw = self.payload_parent_map[payload]
        while True:
            if ppw not in self.pw_parent_map:
                break
            self.pw_parent_map[ppw].pane(ppw, weight=1)
            ppw = self.pw_parent_map[ppw]
        return True

    def calculate_payload_availability(self):
        """Calculate availability of data for display."""
        pla = self.payload_available
        ppm = self.payload_parent_map
        pwpm = self.pw_parent_map
        pwcw = self.pw_current_weights
        pww = self.pw_weights
        for payload in pla:
            pla[payload] = payload.is_payload_available()
            widget = ppm[payload]
            while widget in pwpm:
                pwcw[widget] = 0
                widget = pwpm[widget]
        for key, value in pla.items():
            widget = ppm[key]
            while widget in pwpm:
                if value:
                    pwcw[widget] = pww[widget]
                widget = pwpm[widget]
        if pwcw[ppm[self.base_games]]:
            if pwcw[ppm[self.game_items]]:
                # Same parent tree as base_games so just do child.
                widget = ppm[self.game_games]
                pwcw[widget] = pww[widget]
            if pwcw[ppm[self.repertoire_items]]:
                # Same parent tree as repertoire_items so just do child.
                widget = ppm[self.repertoire_games]
                pwcw[widget] = pww[widget]
            if pwcw[ppm[self.partial_items]]:
                # Same parent tree as partial_items so just do child.
                widget = ppm[self.partial_games]
                pwcw[widget] = pww[widget]

    def configure_panedwindows(self):
        """Display available panedwindows subject to single_view status."""
        for k in self.payload_available:
            widget = self.payload_parent_map[k]
            if self.pw_current_weights[widget]:
                k.insert_payload(widget)
            elif k in widget.panes():
                k.forget_payload(widget)
            while widget in self.pw_parent_map:
                self.pw_parent_map[widget].pane(
                    widget, weight=self.pw_current_weights[widget]
                )
                widget = self.pw_parent_map[widget]

    def set_properties_on_all_game_grids(self, game):
        """Set properties for game on all grids where it is visible."""
        for grid in (
            self.base_games,
            self.game_games,
            self.repertoire_games,
            self.partial_games,
        ):
            grid.set_properties(game)

    def set_bindings_on_item_losing_focus_by_pointer_click(self):
        """Set bindings for an active item when it does not have focus.

        Binary moves of focus between two items by keyboard or popup menu
        action are dealt with elsewhere.

        This method deals with many-to-one possibilities implied by pointer
        click on a specific item.  It could deal with the binary ones too, and
        maybe that is better, or not.

        """
        for item in (
            self.game_items,
            self.repertoire_items,
            self.base_games,
            self.game_games,
            self.repertoire_games,
            self.base_repertoires,
            self.partial_games,
            self.partial_items,
            self.base_partials,
        ):
            if item.bind_for_widget_without_focus():
                break

    def _set_position_analysis_data_source_all_items(self):
        """Set game and repertoire analysis data sources to match database.

        self.ui.database will be None if no database open, or the database
        instance if open.

        """
        for items in (
            self.game_items.order,
            self.repertoire_items.order,
            self.games_and_repertoires_in_toplevels,
        ):
            for item in items:
                item.set_position_analysis_data_source()

    def make_position_analysis_data_source(self):
        """Create a new DataSource for stored chess engine analysis."""
        if self.database:

            # Without the 'is not None' seems unreliable at 08 Nov 2015.
            # What is wrong with 'if <obj>:' where obj is a bsddb3.DB instance?
            # It does work sometimes, so some environment clutter perhaps.
            # Not yet known what happens with sqlite3 and so forth.
            if (
                self.database.is_database_file_active(ANALYSIS_FILE_DEF)
                is not None
            ):
                return self.engineanalysisds(
                    self.database,
                    ANALYSIS_FILE_DEF,
                    VARIATION_FIELD_DEF,
                    newrow=ChessDBrecordAnalysis,
                )
        return None

    def _create_selection_rules_datasource(self, database):
        """Create a new DataSource for list of selection rules."""
        # Probably wrong because datasource = F(active item), but this is a
        # constant at present.
        self.base_selections.set_data_source(
            DataSource(
                database,
                SELECTION_FILE_DEF,
                RULE_FIELD_DEF,
                make_ChessDBrowQuery(self),
            ),
            self.base_selections.on_data_change,
        )
        if self.selection_items.any_items_displayed_of_type():
            self.selection_items.set_insert_or_delete_on_all_items()

    def _show_base_selection_rules(self):
        """Show widget containing list of selection rules in database."""
        if self.selections_pw.pane(self.selection_rules_pw, "weight"):

            # Size is assumed to be (1,1) as created at first call and grid
            # will be filled after <Configure> event to size the widget.
            # Subsequent calls are not associated with a <Configure> event so
            # fill the widget by after_idle(...).
            if self.base_selections.get_frame().winfo_width() != 1:
                self.base_selections.set_partial_key()
                self.selections_pw.after_idle(
                    self.base_selections.load_new_index
                )

    def show_selection_rules_grid(self, database):
        """Show widget containing list of selection rules in database."""
        self._create_selection_rules_datasource(database)

        ## Hack to cope with GUI use while Import in progress.
        # if self.base_selections.get_data_source().get_database() is not None:
        #    self.base_selections.clear_grid_keys()

        self.calculate_payload_availability()
        self.configure_panedwindows()
        self._show_base_selection_rules()
        self.base_selections.set_focus()

    def _set_find_selection_name_games(self, index):
        """Set status text to selection rule name being searched."""
        self.statusbar.set_status_text(
            self.selection_items.get_stack_item(index).get_text_for_statusbar()
        )

    def set_toolbarframe_disabled(self):
        """Set state for widgets in toolbar frame."""
        self.tb_entry.configure(state=tkinter.DISABLED)
        for widget in (self.tb_moveto, self.tb_filter):
            widget.configure(state=tkinter.DISABLED, command="")

    def set_toolbarframe_normal(self, move_to, filter_):
        """Set state for widgets in toolbar frame."""
        self.tb_entry.configure(state=tkinter.NORMAL)
        self.tb_moveto.configure(state=tkinter.NORMAL, command=move_to)
        self.tb_filter.configure(state=tkinter.NORMAL, command=filter_)

    def move_to_game(self):
        """Move to first game with PGN Tag value starting text in filter."""
        self.base_games.move_to_row_in_grid(self.tb_entry.get())

    def move_to_partial(self):
        """Move to first ChessQL statement with name starting text in filter."""
        self.base_partials.move_to_row_in_grid(self.tb_entry.get())

    def move_to_repertoire(self):
        """Move to first repertoire with name starting text in filter."""
        self.base_repertoires.move_to_row_in_grid(self.tb_entry.get())

    def move_to_selection(self):
        """Move to first query with name starting text in filter."""
        self.base_selections.move_to_row_in_grid(self.tb_entry.get())

    def filter_game(self):
        """Show games with PGN Tag value starting text in filter."""
        text = self.tb_entry.get()
        self.base_games.record_count = None
        self.base_games.set_partial_key(text if len(text) else None)

        # Not yet sure if other three filter_* methods need fill_view() call to
        # populate widget without an explicit scroll action by user.
        # '*' seems to be treated as a 0-n wildcard character when passed to
        # SQLite3 via apsw or sqlite3 interfaces, and creates a situation that
        # looks difficult to resolve in DataGrid.
        # (Added much later:)
        # Implementing filters with '*' characters could be done by selection
        # rules evaluated here using the 't' value.
        try:
            self.base_games.fill_view()
        except KeyError:
            if "*" in text:
                tkinter.messagebox.showwarning(
                    parent=self.get_toplevel(),
                    title="Filter Games",
                    message="".join(
                        (
                            "Filter '",
                            text,
                            "' contains a '*'.\n\nThis warning ",
                            "is expected to appear if the database engine is ",
                            "SQLite3 and means an attempt to scroll the list of ",
                            "games will cause a program crash.\n\nA known way ",
                            "to avoid this is menu option 'Select | Game'.\n\n",
                            "Assuming the filter was enabled by menu option ",
                            "'Select | White' the filter can be done by 'Select ",
                            "| Rule', typing 'White starts ",
                            text,
                            "', followed ",
                            "by 'Ctrl + Enter'.\n\nIf the filter was enabled by ",
                            "'Select | Result' the appropriate rule is likely ",
                            "to be 'Result eq *'.",
                        )
                    ),
                )

    def filter_partial(self):
        """Show ChessQL statements with name starting text in filter."""
        text = self.tb_entry.get()
        self.base_partials.load_new_partial_key(text if len(text) else None)

    def filter_repertoire(self):
        """Show repertoires with name starting text in filter."""
        text = self.tb_entry.get()
        self.base_repertoires.load_new_partial_key(text if len(text) else None)

    def filter_selection(self):
        """Show query statements with name starting text in filter."""
        text = self.tb_entry.get()
        self.base_selections.load_new_partial_key(text if len(text) else None)

    def hide_scrollbars(self):
        """Hide the scrollbars in the game display widgets."""
        self.visible_scrollbars = False
        exceptions = []
        for items in (
            self.game_items.order,
            self.repertoire_items.order,
            self.partial_items.order,
            self.selection_items.order,
            self.games_and_repertoires_in_toplevels,
            self.partials_in_toplevels,
            self.selections_in_toplevels,
        ):
            for i in items:
                try:
                    i.hide_scrollbars()
                except tkinter.TclError:
                    exceptions.append((i, items))
        for i, items in exceptions:
            items.remove(i)

    def show_scrollbars(self):
        """Show the scrollbars in the game display widgets."""
        self.visible_scrollbars = True
        exceptions = []
        for items in (
            self.game_items.order,
            self.repertoire_items.order,
            self.partial_items.order,
            self.selection_items.order,
            self.games_and_repertoires_in_toplevels,
            self.partials_in_toplevels,
            self.selections_in_toplevels,
        ):
            for i in items:
                try:
                    i.show_scrollbars()
                except tkinter.TclError:
                    exceptions.append(i, items)
        for i, items in exceptions:
            items.remove(i)

    def export_report(self, result, title):
        """Show a report dialogue if some game scores could not be exported."""
        if result is None:
            tkinter.messagebox.showinfo(
                parent=self.get_toplevel(),
                title=title,
                message="".join(
                    (
                        "No games scores output because none are ",
                        "PGN export format compliant",
                    )
                ),
            )
        elif result is False:
            tkinter.messagebox.showinfo(
                parent=self.get_toplevel(),
                title=title,
                message="".join(
                    (
                        "Some games scores are not output because they are ",
                        "not PGN export format compliant",
                    )
                ),
            )



# view_players_bsddb3.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Print names in games_Black and games_White records where the name with
',' and '.' ripped out is not the name in the record."""


if __name__ == "__main__":

    import os
    import ast
    import re

    re_normalize_player_name = re.compile("([^,\.\s]+(?:,|\.)?)(?:\s*)")

    file_dbname_map = {
        "games_Black": "Black",
        "games_White": "White",
    }

    # bsddb removed from Python 3.n
    try:
        from bsddb3.db import DB
    except ImportError:
        from bsddb.db import DB

    import tkinter.filedialog

    directoryname = tkinter.filedialog.askdirectory(title="DB file to view")
    if directoryname:
        for bn, tn in file_dbname_map.items():
            db = DB()
            db.open(
                os.path.join(directoryname, os.path.basename(directoryname)),
                bn,
            )
            c = db.cursor()
            rec = c.first()
            print(tn)
            while rec:
                k, v = rec
                ks = k.decode("iso-8859-1")
                nks = " ".join(re_normalize_player_name.findall(ks))
                if ks != nks:
                    print(ks)
                rec = c.next()
                if rec is None:
                    break
                # print(rec)
            c.close()
            db.close()
            del db



# chessdb.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using Berkeley DB.

The Berkeley DB interface has significantly worse performance than DPT when
doing multi-index searches.  However it is retained since DPT became available
because tracking down problems in the chess logic using IDLE can be easier
in the *nix environment.
"""

import os

from bsddb3.db import (
    DB_BTREE,
    DB_HASH,
    DB_RECNO,
    DB_DUPSORT,
    DB_DUP,
    DB_CREATE,
    DB_RECOVER,
    DB_INIT_MPOOL,
    DB_INIT_LOCK,
    DB_INIT_LOG,
    DB_INIT_TXN,
    DB_PRIVATE,
)

from solentware_base import bsddb3_database
from solentware_base.core.constants import (
    SUBFILE_DELIMITER,
    EXISTENCE_BITMAP_SUFFIX,
    SEGMENT_SUFFIX,
    SECONDARY,
)

from ..core.filespec import (
    FileSpec,
    DB_ENVIRONMENT_GIGABYTES,
    DB_ENVIRONMENT_BYTES,
    DB_ENVIRONMENT_MAXLOCKS,
)
from ..basecore import database


class ChessDatabase(database.Database, bsddb3_database.Database):

    """Provide access to a database of games of chess."""

    _deferred_update_process = os.path.join(
        os.path.basename(os.path.dirname(__file__)), "runchessdbdu.py"
    )

    def __init__(self, DBfile, **kargs):
        """Define chess database.

        **kargs
        Arguments are passed through to superclass __init__.

        """
        dbnames = FileSpec(**kargs)

        environment = {
            "flags": (
                DB_CREATE
                | DB_RECOVER
                | DB_INIT_MPOOL
                | DB_INIT_LOCK
                | DB_INIT_LOG
                | DB_INIT_TXN
                | DB_PRIVATE
            ),
            "gbytes": DB_ENVIRONMENT_GIGABYTES,
            "bytes": DB_ENVIRONMENT_BYTES,
            "maxlocks": DB_ENVIRONMENT_MAXLOCKS,
        }

        super().__init__(dbnames, DBfile, environment)

    def delete_database(self):
        """Close and delete the open chess database."""
        return super().delete_database(
            (self.database_file, self.dbenv.get_lg_dir().decode())
        )



# gamerow.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Create widgets that display tag roster details of games on database.
"""

import tkinter

from solentware_grid.gui.datarow import (
    GRID_COLUMNCONFIGURE,
    GRID_CONFIGURE,
    WIDGET_CONFIGURE,
    WIDGET,
    ROW,
)

from pgn_read.core.constants import (
    TAG_WHITE,
    TAG_BLACK,
    TAG_RESULT,
    TAG_EVENT,
    TAG_DATE,
    SEVEN_TAG_ROSTER,
    DEFAULT_TAG_VALUE,
    DEFAULT_TAG_DATE_VALUE,
    DEFAULT_TAG_RESULT_VALUE,
)

from .datarow import DataRow
from ..core.chessrecord import ChessDBrecordGameTags
from .gamedbedit import GameDbEdit
from .gamedbdelete import GameDbDelete
from .gamedbshow import GameDbShow
from . import constants

ON_DISPLAY_COLOUR = "#eba610"  # a pale orange


class ChessDBrowGame(ChessDBrecordGameTags, DataRow):
    """Define row in list of games.

    Add row methods to the chess game record definition.

    """

    header_specification = [
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_WHITE,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="player"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_RESULT,
                anchor=tkinter.CENTER,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=1, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="score"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_BLACK,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=2, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="player"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_EVENT,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=3, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="event"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text=TAG_DATE,
                anchor=tkinter.W,
                padx=0,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=4, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=1, uniform="date"),
            ROW: 0,
        },
        {
            WIDGET: tkinter.Label,
            WIDGET_CONFIGURE: dict(
                text="Tags",
                anchor=tkinter.W,
                padx=10,
                pady=1,
                font="TkDefaultFont",
            ),
            GRID_CONFIGURE: dict(column=5, sticky=tkinter.EW),
            GRID_COLUMNCONFIGURE: dict(weight=4, uniform="tags"),
            ROW: 0,
        },
    ]

    def __init__(self, database=None, ui=None):
        """Extend and associate record definition with database.

        database - the open database that is source of row data
        ui - the ChessUI instamce

        """
        super(ChessDBrowGame, self).__init__()
        self.ui = ui
        self.set_database(database)
        self.row_specification = [
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=0, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.CENTER,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=1, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=2, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=3, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=0,
                ),
                GRID_CONFIGURE: dict(column=4, sticky=tkinter.EW),
                ROW: 0,
            },
            {
                WIDGET: tkinter.Label,
                WIDGET_CONFIGURE: dict(
                    anchor=tkinter.W,
                    font=constants.LISTS_OF_GAMES_FONT,
                    pady=1,
                    padx=10,
                ),
                GRID_CONFIGURE: dict(column=5, sticky=tkinter.EW),
                ROW: 0,
            },
        ]

    def show_row(self, dialog, oldobject):
        """Return a GameDbShow toplevel for instance.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return GameDbShow(dialog, oldobject, ui=self.ui)

    def delete_row(self, dialog, oldobject):
        """Return a GameDbDelete toplevel for instance.

        dialog - a Toplevel
        oldobject - a ChessDBrecordGame containing original data

        """
        return GameDbDelete(dialog, oldobject, ui=self.ui)

    def edit_row(self, dialog, newobject, oldobject, showinitial=True):
        """Return a GameDbEdit toplevel for instance.

        dialog - a Toplevel
        newobject - a ChessDBrecordGame containing original data to be edited
        oldobject - a ChessDBrecordGame containing original data
        showintial == True - show both original and edited data

        """
        return GameDbEdit(
            newobject, dialog, oldobject, showinitial=showinitial, ui=self.ui
        )

    def grid_row(self, **kargs):
        """Return super(ChessDBrowGame,).grid_row(textitems=(...), **kargs).

        Create textitems argument for ChessDBrowGame instance.

        """
        tags = self.value.collected_game._tags
        return super(ChessDBrowGame, self).grid_row(
            textitems=(
                tags.get(TAG_WHITE, DEFAULT_TAG_VALUE),
                tags.get(TAG_RESULT, DEFAULT_TAG_RESULT_VALUE),
                tags.get(TAG_BLACK, DEFAULT_TAG_VALUE),
                tags.get(TAG_EVENT, DEFAULT_TAG_VALUE),
                tags.get(TAG_DATE, DEFAULT_TAG_DATE_VALUE),
                "  ".join(
                    [
                        "".join((tag, ' "', value, '"'))
                        for tag, value in self.get_tags_display_order(
                            self.value
                        )
                    ]
                ),
            ),
            **kargs
        )

    def get_tags_display_order(self, pgn):
        str_tags = []
        other_tags = []
        tags = self.value.collected_game._tags
        for t in SEVEN_TAG_ROSTER:
            if t not in constants.GRID_HEADER_SEVEN_TAG_ROSTER:
                str_tags.append((t, tags.get(t, DEFAULT_TAG_VALUE)))
        for t, v in sorted(tags.items()):
            if t not in SEVEN_TAG_ROSTER:
                other_tags.append((t, v))
        return str_tags + other_tags

    def set_background_on_display(self, widgets):
        self._current_row_background = ON_DISPLAY_COLOUR
        self.set_background(widgets, self._current_row_background)

    def grid_row_on_display(self, **kargs):
        self._current_row_background = ON_DISPLAY_COLOUR
        return self.grid_row(background=ON_DISPLAY_COLOUR, **kargs)


def make_ChessDBrowGame(chessui):
    """Make ChessDBrowGame with reference to ChessUI instance"""

    def make_position(database=None):
        return ChessDBrowGame(database=database, ui=chessui)

    return make_position



# gametoplevel.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Toplevel widgets to display and edit game scores.

These two classes display games in their own Toplevel widget: they are used
in the gamedbdelete, gamedbedit, and gamedbshow, modules.

"""
from .game import Game
from .gameedit import GameEdit
from .toplevelpgn import ToplevelPGN


class GameToplevel(ToplevelPGN, Game):

    """Customize Game to be the single instance in a Toplevel widget."""


class GameToplevelEdit(ToplevelPGN, GameEdit):

    """Customize GameEdit to be the single instance in a Toplevel widget."""

    def create_primary_activity_popup(self):
        popup = super().create_primary_activity_popup()
        self.add_pgn_navigation_to_submenu_of_popup(
            popup, index=self.analyse_popup_label
        )
        self.add_pgn_insert_to_submenu_of_popup(
            popup,
            include_ooo=True,
            include_move_rav=True,
            index=self.analyse_popup_label,
        )
        return popup



# chessvedis.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database using vedis.
"""

import os

from solentware_base import vedis_database
from solentware_base.core.constants import (
    FILEDESC,
)

from ..core.filespec import FileSpec
from ..basecore import database


class ChessDatabaseError(Exception):
    pass


class ChessDatabase(database.Database, vedis_database.Database):

    """Provide access to a database of games of chess."""

    _deferred_update_process = os.path.join(
        os.path.basename(os.path.dirname(__file__)), "runchessvedisdu.py"
    )

    def __init__(self, nosqlfile, **kargs):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        names = FileSpec(**kargs)

        if not kargs.get("allowcreate", False):
            try:
                for t in names:
                    if FILEDESC in names[t]:
                        del names[t][FILEDESC]
            except:
                if __name__ == "__main__":
                    raise
                else:
                    raise ChessDatabaseError("vedis description invalid")

        try:
            super().__init__(names, nosqlfile, **kargs)
        except ChessDatabaseError:
            if __name__ == "__main__":
                raise
            else:
                raise ChessDatabaseError("vedis description invalid")

    def delete_database(self):
        """Close and delete the open chess database."""
        return super().delete_database((self.database_file,))



"""User interface to a chess game database.

Module "chess" is the top-level module.

Module "chess_ui" controls the User Interface.

Module "chessdu" is the top-level module in sub-processes started to import
data from PGN files.

Four sets of modules provide the user interface for games, repertoires, partial
positions, and selection rules.  Some repertoire classes are defined in the
corresponding game module rather than a repertoire module.  The absence of
modules called "repertoire", "repertoireedit", and "repertoiregrid", indicates
which ones.

The remaining modules provide supporting facilities.
"""



# cqldbshow.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise show toplevel to display Chess Query Language (ChessQL) statement
record.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

"""

from solentware_grid.gui.datashow import DataShow

from solentware_misc.gui.exceptionhandler import ExceptionHandler

from .cqltoplevel import CQLToplevel
from .topleveltext import ShowText


class CQLDbShow(ExceptionHandler, ShowText, DataShow):
    """Show ChessQL statement from database.

    parent is used as the master argument in a CQLToplevel call.

    ui is used as the ui argument in a CQLToplevel call.

    parent, oldobject, and the CQLToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods get_title_for_object and set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "ChessQL Statement"

    def __init__(self, parent, oldobject, ui=None):
        """Extend and create toplevel widget to display ChessQL statement."""
        # Toplevel title set '' in __init__ and to proper value in initialize.
        super().__init__(
            oldobject, parent, CQLToplevel(master=parent, ui=ui), ""
        )
        self.initialize()

    def get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a ChessQL statement object_.

        Default value of object_ is object attribute from DataShow class.

        """
        if object_ is None:
            object_ = self.object
        return "  ".join(
            (
                self.text_name.join(("Show ", ":")),
                object_.value.get_name_text(),
            )
        )

    @property
    def ui_base_table(self):
        return self.ui.base_partials

    @property
    def ui_items_in_toplevels(self):
        return self.ui.partials_in_toplevels

    @property
    def ui(self):
        return self.oldview.ui

    def set_item(self, view, object_):
        view.cql_statement.process_statement(object_.get_srvalue())
        view.set_and_tag_item_text()



"""Data manipulation for chess game data and access to databases.

The data access methods are independent from any particular database engine,
but the design assumes the "key:value" structure used in Berkeley DB.

The sibling packages apsw, db, dpt, and sqlite, use the methods provided by the
solentware_base package interfaces to Berkeley DB, DPT, and Sqlite3, to
implement the "key:value" structure.
"""



# database.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""ChessTab database methods common to all database engine interfaces.
"""

import os
import bz2
import shutil

from ..core.filespec import (
    NEWGAMES_FIELD_DEF,
    NEWGAMES_FIELD_VALUE,
    PARTIAL_FILE_DEF,
)
from .. import APPLICATION_NAME, ERROR_LOG


class Database:

    """ """

    def use_deferred_update_process(self, **kargs):
        """Return path to deferred update module.

        **kargs - soak up any arguments other database engines need.

        """
        return self._deferred_update_process

    def open_database(self, files=None):
        """Return True to fit behaviour of dpt version of this method."""
        super().open_database(files=files)
        return True

    def dump_database(self, names=()):
        """Dump database in compressed files."""
        for n in names:
            c = bz2.BZ2Compressor()
            archivename = ".".join((n, "broken", "bz2"))
            fi = open(n, "rb")
            fo = open(archivename, "wb")
            try:
                inp = fi.read(10000000)
                while inp:
                    co = c.compress(inp)
                    if co:
                        fo.write(co)
                    inp = fi.read(10000000)
                co = c.flush()
                if co:
                    fo.write(co)
            finally:
                fo.close()
                fi.close()

    def delete_backups(self, names=()):
        """Delete backup files."""
        for n in names:
            archiveguard = ".".join((n, "grd"))
            archivename = ".".join((n, "bz2"))
            try:
                os.remove(archiveguard)
            except:
                pass
            try:
                os.remove(archivename)
            except:
                pass

    def restore_backups(self, names=()):
        """Restore database from backup files."""
        for n in names:
            c = bz2.BZ2Decompressor()
            archivename = ".".join((n, "bz2"))
            fi = open(archivename, "rb")
            fo = open(n, "wb")
            try:
                inp = fi.read(1000000)
                while inp:
                    co = c.decompress(inp)
                    if co:
                        fo.write(co)
                    inp = fi.read(1000000)
            finally:
                fo.close()
                fi.close()
        return True

    def delete_database(self, names):
        """Close and delete the open chess database."""
        listnames = set(n for n in os.listdir(self.home_directory))
        homenames = set(n for n in names if os.path.basename(n) in listnames)
        if ERROR_LOG in listnames:
            homenames.add(os.path.join(self.home_directory, ERROR_LOG))
        if len(listnames - set(os.path.basename(h) for h in homenames)):
            message = "".join(
                (
                    "There is at least one file or folder in\n\n",
                    self.home_directory,
                    "\n\nwhich may not be part of the database.  These items ",
                    "have not been deleted by ",
                    APPLICATION_NAME,
                    ".",
                )
            )
        else:
            message = None
        self.close_database()
        for h in homenames:
            if os.path.isdir(h):
                shutil.rmtree(h, ignore_errors=True)
            else:
                os.remove(h)
        try:
            os.rmdir(self.home_directory)
        except:
            pass
        return message

    def get_archive_names(self, files=()):
        """Return names and operating system files for archives and guards"""
        names = (self.database_file,)
        archives = dict()
        guards = dict()
        for n in names:
            archiveguard = ".".join((n, "grd"))
            archivefile = ".".join((n, "bz2"))
            for d, f in ((archives, archivefile), (guards, archiveguard)):
                if os.path.exists(f):
                    d[n] = f
        return (names, archives, guards)

    def open_after_import_without_backups(self, **ka):
        """Return True after doing database engine specific open actions.

        For SQLite3 and Berkeley DB just call open_database.

        """
        super().open_database()

        # Return True to fit behaviour of chessdpt module version of method.
        return True

    def open_after_import_with_backups(self, files=()):
        """Return True after doing database engine specific open actions.

        For SQLite3 and Berkeley DB just call open_database.

        """
        super().open_database()

        # Return True to fit behaviour of chessdpt module version of method.
        return True

    def save_broken_database_details(self, files=()):
        """Save database engine specific detail of broken files to be restored.

        It is assumed that the Database Services object exists.

        """
        pass

    def adjust_database_for_retry_import(self, files):
        """Database engine specific actions to do before re-trying an import"""
        pass

    def mark_partial_positions_to_be_recalculated(self):
        """File all partial positions to be recalculated."""
        self.start_transaction()
        allrecords = self.recordlist_ebm(PARTIAL_FILE_DEF)
        self.file_records_under(
            PARTIAL_FILE_DEF,
            NEWGAMES_FIELD_DEF,
            allrecords,
            self.encode_record_selector(NEWGAMES_FIELD_VALUE),
        )
        allrecords.close()
        self.commit()



# uci.py
# Copyright 2015 Roger Marsh
# License: See LICENSE.TXT (BSD licence)

"""Collate position analysis from multiple chess engines.

Chess engines must support the UCI (Universal Chess Interface).

The UCI commands relevant to analysing a single position are supported.

The commands related to 'infinite analysis' are not supported because this
module expects the chess engine to finish a command and give a response without
the need to interrupt it.

"""

import multiprocessing
from queue import Empty
import os

from solentware_grid.gui.dataedit import RecordEdit

from uci_net.uci_driver_over_tcp import UCIDriverOverTCP
from uci_net.engine import (
    Engine,
    EngineInterface,
    CommandsFromEngine,
    CommandsToEngine,
    GoSubCommands,
    PositionSubCommands,
    OptionParameters,
    ReservedOptionNames,
    SetoptionSubCommands,
    InfoParameters,
    ScoreInfoValueNames,
)

from .uci_to_pgn import generate_pgn_for_uci_moves_in_position
from .chessrecord import ChessDBrecordAnalysis
from .analysis import Analysis


class UCI(object):
    """Control multiple chess engines using the UCI."""

    def __init__(self):
        """Set data stuctures to control multiple UCI chess engines."""
        self.engine_counter = 0
        self.uci_drivers_index = dict()

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # try:
        #    self.uci_drivers_reply = multiprocessing.Queue()
        # except OSError:
        #    self.uci_drivers_reply = None
        self.uci_drivers_reply = multiprocessing.Queue()

        self.uci_drivers = dict()
        self.uci_drivers_fen = dict()
        self.uci_active_engines = dict()
        self.ui_analysis_queue = None
        self.position_analysis = dict()
        self._use_ucinewgame = True
        self._clear_hash = True
        self._go_depth = 10

        # 0 (zero) means use the default assumed by the engine.
        self._multipv = 0
        self._hash_size = 0

        self.analysis_record = ChessDBrecordAnalysis()
        self.analysis_data_source = None

        # These are new and will replace some of above attributes.
        self.positions_pending = dict()
        self.engine_queues = dict()
        self.clear_hash_after_bestmove = dict()
        self.set_option_on_empty_queues = set()

    def kill_engine(self, number):
        """Attempt to kill engine and return True if successful."""
        ui_name = self.uci_drivers_index[number]
        ei = self.uci_drivers[ui_name]
        try:
            ei.to_driver_queue.put(CommandsToEngine.quit_)
        except:
            return False
        ei.driver.join()
        engine_name = ei.parser.name
        if engine_name in self.uci_active_engines:

            # The command 'stockfish argument' starts stockfish but the engine
            # quits immediately citing 'Unknown command: argument'.
            # These engine_name entries do not exist in this case.
            self.uci_active_engines[engine_name].discard(ui_name)
            if not self.uci_active_engines[engine_name]:
                del self.uci_active_engines[engine_name]
                del self.positions_pending[engine_name]
                del self.engine_queues[engine_name]

        del self.clear_hash_after_bestmove[ui_name]
        del self.uci_drivers[ui_name]
        del self.uci_drivers_fen[ui_name]
        del self.uci_drivers_index[number]
        return True

    def quit_all_engines(self):
        """Attempt to kill all engines and return list of those not killed."""
        joiners = []
        non_joiners = []
        for ui_name, ei in self.uci_drivers.items():
            try:
                ei.to_driver_queue.put(CommandsToEngine.quit_)
            except:
                non_joiners.append((ui_name, ei.driver.pid))
                continue
            joiners.append(ei.driver)
        for j in joiners:
            j.join()
        self.uci_active_engines.clear()
        self.positions_pending.clear()
        self.engine_queues.clear()
        self.uci_drivers.clear()
        self.uci_drivers_index.clear()
        self.uci_drivers_fen.clear()
        self.clear_hash_after_bestmove.clear()
        return non_joiners

    def run_engine(self, program_file_name, args):
        """Start a process to run an UCI chess engine."""

        # Avoid "OSError: [WinError 535] Pipe connected"  at Python3.3 running
        # under Wine on FreeBSD 10.1 by disabling the UCI functions.
        # Assume all later Pythons are affected because they do not install
        # under Wine at time of writing.
        # The OSError stopped happening by wine-2.0_3,1 on FreeBSD 10.1 but
        # get_nowait() fails to 'not wait', so ChessTab never gets going under
        # wine at present.  Leave alone because it looks like the problem is
        # being shifted constructively.
        # At Python3.5 running under Wine on FreeBSD 10.1, get() does not wait
        # when the queue is empty either, and ChessTab does not run under
        # Python3.3 because it uses asyncio: so no point in disabling.
        # if self.uci_drivers_reply is None:
        #    return

        self.engine_counter += 1
        ui_name = ": ".join(
            (
                str(self.engine_counter),
                os.path.splitext(os.path.basename(program_file_name))[0],
            )
        )
        to_driver_queue = multiprocessing.Queue()
        driver = multiprocessing.Process(
            target=run_driver,
            args=(
                to_driver_queue,
                self.uci_drivers_reply,
                program_file_name,
                args,
                ui_name,
            ),
        )
        driver.start()
        self.uci_drivers[ui_name] = EngineInterface(
            driver=driver,
            program_file_name=program_file_name,
            to_driver_queue=to_driver_queue,
            parser=Engine(),
        )
        self.uci_drivers_index[self.engine_counter] = ui_name
        self.uci_drivers_fen[ui_name] = None
        self.clear_hash_after_bestmove[ui_name] = False

        # Send the initial 'uci' command to the engine
        to_driver_queue.put(CommandsToEngine.uci)

    def get_engine_responses(self):
        """Drain the queue of responses and process them.

        Calls of this method are scheduled by tkinter.after().

        """
        while True:
            try:
                item = self.uci_drivers_reply.get_nowait()
            except Empty:
                break
            try:
                self.process_commands_from_engines(item)
            except:
                pass

    def process_commands_from_engines(self, response):
        """Process responses from UCI chess engines.

        Calls of this method are scheduled by tkinter.after().

        """
        ui_name, commands = response

        # UCI commands are generally ignored if not understood.
        # Follow this principle here.
        try:
            self.uci_drivers[ui_name].parser.process_engine_commands(response)
        except:
            return

        c = commands[-1].split(None, maxsplit=1)[0]
        if c == CommandsFromEngine.bestmove:
            self.bestmove(ui_name)
        elif c == CommandsFromEngine.readyok:
            self.readyok(ui_name)
        elif c == CommandsFromEngine.uciok:
            self.uciok(ui_name)
        elif c == CommandsFromEngine.registration:
            self.registration(ui_name)
        elif c == CommandsFromEngine.copyprotection:
            self.copyprotection(ui_name)
        else:
            pass

    def get_analysis_requests(self):
        """Add requests for analysis of positions to engine queues."""
        if (
            self.set_option_on_empty_queues
            and self.is_positions_pending_empty()
        ):
            if ReservedOptionNames.Hash in self.set_option_on_empty_queues:

                # Postpone Hash action until all engines ready.
                for ei in self.uci_drivers.values():
                    eip = ei.parser
                    if eip.readyok_expected or eip.uciok_expected is not False:
                        return

                for ei in self.uci_drivers.values():
                    eip = ei.parser
                    eipoh = eip.options.get(ReservedOptionNames.Hash)
                    if eipoh is None:
                        continue
                    if not self.hash_size:
                        newhash = eipoh[1].get(OptionParameters.default)
                        if newhash is None:
                            continue
                    else:
                        min_ = eipoh[1].get(OptionParameters.min_)
                        max_ = eipoh[1].get(OptionParameters.max_)
                        if min_ is None or max_ is None:
                            continue
                        newhash = max(
                            min(self.hash_size, int(max_)), int(min_)
                        )
                    ei.to_driver_queue.put(
                        " ".join(
                            (
                                CommandsToEngine.setoption,
                                SetoptionSubCommands.name,
                                ReservedOptionNames.Hash,
                                SetoptionSubCommands.value,
                                str(newhash),
                            )
                        )
                    )
                    ei.to_driver_queue.put(CommandsToEngine.isready)
                    ei.parser.readyok_expected = True
                self.set_option_on_empty_queues.discard(
                    ReservedOptionNames.Hash
                )
            return

        requests = {}
        engine_queues = self.engine_queues
        positions_pending = self.positions_pending
        while True:
            try:
                item = self.ui_analysis_queue.get_nowait()
            except Empty:
                break
            except:
                if self.ui_analysis_queue is None:
                    break
                raise
            game, position_data = item
            if game is position_data:
                continue
            requests.setdefault(game, []).append(position_data)
            for v in engine_queues.values():
                if game in v:
                    v.append(v.pop(v.index(game)))
                else:
                    v.append(game)
        for v in positions_pending.values():
            v.update({rk: rv.copy() for rk, rv in requests.items()})

        uci_drivers_fen = self.uci_drivers_fen
        for k, ei in self.uci_drivers.items():
            if uci_drivers_fen[k] is not None:
                continue
            eip = ei.parser
            if eip.readyok_expected:
                continue
            if eip.uciok_expected is not False:
                continue
            eq = engine_queues[eip.name]
            epp = positions_pending[eip.name]
            while len(eq):
                if len(epp[eq[-1]]):
                    position_data = epp[eq[-1]].pop()
                    if self.process_analysis_request(ei, position_data):
                        uci_drivers_fen[k] = position_data.position
                        break
                else:
                    del epp[eq[-1]]
                    del eq[-1]

    def set_analysis_queue(self, queue):
        """Note queue for return of analysis by UCI engine to user interface."""
        self.ui_analysis_queue = queue

    def process_analysis_request(self, ei, position_data):
        """Find existing analysis or ask engine for deeper or wider analysis."""
        engine_name = ei.parser.name

        # (0, 0) or value read from database by requestor.
        depth, multipv = position_data.scale.get(engine_name, (0, 0))
        if depth >= self._go_depth and multipv >= self._multipv:
            return False

        # Is analysis available already from non-database store request?
        fen = position_data.position
        position_analysis_fen = self.position_analysis.get(fen)
        if position_analysis_fen:
            if engine_name in position_analysis_fen.scale:
                depth, multipv = position_analysis_fen.scale[engine_name]
                if depth >= self._go_depth and multipv >= self._multipv:
                    return False

        # Can the engine cope with the multiPV option value?
        ronmpv = ReservedOptionNames.MultiPV
        udpo = ei.parser.options
        if multipv < self._multipv:
            if ronmpv not in udpo:
                if depth >= self._go_depth:
                    return False
                multipv = 1
            else:
                mpv = min(
                    max(
                        int(udpo[ronmpv][1][OptionParameters.min_]),
                        self._multipv,
                    ),
                    int(udpo[ronmpv][1][OptionParameters.max_]),
                )
                if depth >= self._go_depth and multipv >= mpv:
                    return False
                if mpv < multipv:
                    multipv = mpv

        # Put commands to do analysis on engine command queue.
        # If analysis already exists do not reduce depth; multipv is already
        # as high as possible given requested value.
        if ronmpv in udpo:
            ei.to_driver_queue.put(
                " ".join(
                    (
                        CommandsToEngine.setoption,
                        SetoptionSubCommands.name,
                        ronmpv,
                        SetoptionSubCommands.value,
                        str(max(multipv, self._multipv)),
                    )
                )
            )
        ei.to_driver_queue.put(
            " ".join((CommandsToEngine.position, PositionSubCommands.fen, fen))
        )
        ei.to_driver_queue.put(
            " ".join(
                (
                    CommandsToEngine.go,
                    GoSubCommands.depth,
                    str(max(self.go_depth, depth)),
                )
            )
        )
        return True

    def uciok(self, ui_name):
        """Process uciok command from UCI chess engine."""
        ei = self.uci_drivers[ui_name]
        eip = ei.parser
        eip.uciok_expected = False
        self.uci_active_engines.setdefault(eip.name, set()).add(ui_name)
        self.positions_pending.setdefault(eip.name, {})
        self.engine_queues.setdefault(eip.name, [])
        if self.use_ucinewgame:
            ei.to_driver_queue.put(CommandsToEngine.ucinewgame)
            ei.to_driver_queue.put(CommandsToEngine.isready)
            eip.readyok_expected = True

    def readyok(self, ui_name):
        """Process readyok command from UCI chess engine."""
        ei = self.uci_drivers[ui_name]
        if self.clear_hash_after_bestmove[ui_name]:
            ei.to_driver_queue.put(CommandsToEngine.isready)

            # Assume all engines have the 'Clear Hash' option.
            # Partly to avoid changing uci package right now to ask.
            ei.to_driver_queue.put(
                " ".join(
                    (
                        CommandsToEngine.setoption,
                        SetoptionSubCommands.name,
                        ReservedOptionNames.clear_hash,
                    )
                )
            )

            self.clear_hash_after_bestmove[ui_name] = False
            return
        ei.parser.readyok_expected = False

    def bestmove(self, ui_name):
        """Extract engine analysis and tidy up engine.

        This is where the analysis snapshots held in the Engine object should
        be pruned if required: after the extraction.

        """
        self.extract_variations_from_engine_bestmove(ui_name)
        self.uci_drivers_fen[ui_name] = None
        if self.use_ucinewgame:
            ei = self.uci_drivers[ui_name]
            ei.to_driver_queue.put(CommandsToEngine.ucinewgame)
            ei.to_driver_queue.put(CommandsToEngine.isready)
            ei.parser.readyok_expected = True
        if self.clear_hash:
            self.clear_hash_after_bestmove[ui_name] = True

    def copyprotection(self, ui_name):
        """ """
        print(
            CommandsFromEngine.copyprotection,
            self.uci_drivers[ui_name].parser.copyprotection,
        )

    def registration(self, ui_name):
        """ """
        print(
            CommandsFromEngine.registration,
            self.uci_drivers[ui_name].parser.registration,
        )

    @property
    def use_ucinewgame(self):
        """True if ucinewgame command to be used before each move."""
        return self._use_ucinewgame

    @use_ucinewgame.setter
    def use_ucinewgame(self, value):
        """Set to use, or not use, before every position command."""
        self._use_ucinewgame = bool(value)

    @property
    def clear_hash(self):
        """True if clear hash command to be used before each move."""
        return self._clear_hash

    @clear_hash.setter
    def clear_hash(self, value):
        """Set to use, or not use, before every position command."""
        self._clear_hash = bool(value)

    @property
    def go_depth(self):
        """The depth value to use in go depth commands."""
        return self._go_depth

    @go_depth.setter
    def go_depth(self, value):
        """Set to the depth used in all go commands."""
        self._go_depth = value

    @property
    def multipv(self):
        """The value to use in setoption MultiPV commands."""
        return self._multipv

    @multipv.setter
    def multipv(self, value):
        """Set to the value used in all setoption MultiPV commands."""
        self._multipv = value

    @property
    def hash_size(self):
        """The value to use in setoption hash commands."""
        return self._hash_size

    @hash_size.setter
    def hash_size(self, value):
        """Set to the value used in all setoption hash commands."""
        self._hash_size = value

    def any_readyok_pending(self):
        """Return True if expected engine readyok commands not yet received."""
        for ei in self.uci_drivers.values():
            if ei.parser.readyok_expected:
                return True
        return False

    def extract_variations_from_engine_bestmove(self, ui_name):
        """Extract and save on database new wider or deeper analysis."""
        engine_parser = self.uci_drivers[ui_name].parser
        engine_name = engine_parser.name
        snapshot = engine_parser.info[-1][-1]
        multipv = snapshot.multipv
        fen = self.uci_drivers_fen[ui_name]

        # Invert the sorted list comprehension to allow for different content
        # in chess engine 'score' info when mate, or checkmate and stalemate
        # positions, are involved.
        score = InfoParameters.score
        cp = ScoreInfoValueNames.cp
        mate = ScoreInfoValueNames.mate
        lines = []
        for k, v in snapshot.pv_group.items():
            lines.append(
                [
                    int(k) if k else 0,
                    None,
                    v[InfoParameters.depth],
                    generate_pgn_for_uci_moves_in_position(
                        v[InfoParameters.pv][0], fen
                    ),
                ]
            )
            if cp in v[score]:
                lines[-1][1] = v[score][cp]
            elif mate in v[score]:
                lines[-1][1] = v[score][mate] + "0000"
            else:
                lines[-1][1] = 0
        lines = sorted(lines)[: int(multipv) if multipv is not None else 1]

        a = Analysis()
        a.scale = {engine_name: (min([int(d[2]) for d in lines]), len(lines))}
        a.position = fen
        a.variations = {engine_name: [(d[1], d[3]) for d in lines]}

        # The write to analysis file will be done here.
        asd = self.analysis_data_source
        if asd:
            records = asd.get_position_analysis_records(a.position)
            for r in records:
                if engine_name in r.value.scale:
                    er = r.clone()
                    er.value.scale.update(a.scale)
                    er.value.variations.update(a.variations)
                    inserter = RecordEdit(er, r)
                    inserter.set_data_source(asd, inserter.on_data_change)
                    inserter.edit()
                    break
            else:
                if records:
                    r = records[-1]
                    er = r.clone()
                    er.value.scale.update(a.scale)
                    er.value.variations.update(a.variations)
                    inserter = RecordEdit(er, r)
                    inserter.set_data_source(asd, inserter.on_data_change)
                    inserter.edit()
                else:
                    v = self.analysis_record.value
                    v.empty()
                    v.scale = a.scale
                    v.position = a.position
                    v.variations = a.variations
                    inserter = RecordEdit(self.analysis_record, None)
                    inserter.set_data_source(asd, None)
                    self.analysis_record.set_database(
                        inserter.get_data_source().dbhome
                    )
                    self.analysis_record.key.recno = None  # 0
                    inserter.put()
        else:
            pa = self.position_analysis.setdefault(a.position, a)
            if pa is not a:
                pa.scale.update(a.scale)
                pa.variations.update(a.variations)

    def set_position_analysis_data_source(self, datasource=None):
        """Link database analysis reader to open database."""
        self.analysis_data_source = datasource
        if datasource is None:
            self.analysis_record.set_database(None)
        else:
            self.analysis_record.set_database(datasource.get_database())

    def is_positions_pending_empty(self):
        """Return True if no positions are in the queue for analysis."""
        for pending in self.positions_pending.values():
            for p in pending.values():
                if len(p):
                    return False
        return True


def run_driver(to_driver_queue, to_ui_queue, path, args, ui_name):
    """Start UCI chess engine and enter loop sending queued resuests to engine."""
    driver = UCIDriverOverTCP(to_ui_queue, ui_name)
    try:
        driver.start_engine(path, args)
    except:
        to_ui_queue.put(("start failed", (ui_name,)))
        return
    to_ui_queue.put(("started", (ui_name,)))
    while True:
        command = to_driver_queue.get()
        if command == CommandsToEngine.quit_:
            break
        driver.send_to_engine(command)
    driver.quit_engine()



# constants.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""User Interface constants.

Many of the constants are used to generate tkinter.Text tag and mark names
used when highlighting game scores in various widgets.

Assume these fonts have been downloaded and installed:

Chess Cases, Chess Lucena, Chess Merida, and Chess Motif fonts
by Armando H Marroquin.

At time of writing available from:

www.enpassant.dk/chess/fonteng.htm
"""

from pgn_read.core.constants import (
    SEVEN_TAG_ROSTER,
    TAG_WHITE,
    TAG_BLACK,
    TAG_RESULT,
    TAG_EVENT,
    TAG_DATE,
)

from ..core.constants import (
    REPERTOIRE_TAG_ORDER,
    UNKNOWN_RESULT,
)

# names of chess board fonts in decreasing preference order.
# The 'win32' order is my preference, but Merida and Motif are microsoft-symbol
# fonts according to description in /usr/local/share/fonts/../fonts.dir after
# installation on FreeBSD 10.1 and produce the default characters, not pieces.
import sys

if sys.platform == "win32":
    PREFERRED_PIECES = (
        "Chess Merida",
        "Chess Cases",
        "Chess Motif",
        "Chess Lucena",
    )
else:
    PREFERRED_PIECES = (
        "Chess Cases",
        "Chess Lucena",
        "Chess Merida",
        "Chess Motif",
    )
del sys

# names of Tk named fonts.
MOVES_PLAYED_IN_GAME_FONT = "moves"
PIECES_ON_BOARD_FONT = "pieces"
WILDPIECES_ON_BOARD_FONT = "wildpieces"
LISTS_OF_GAMES_FONT = "lists"
TAGS_VARIATIONS_COMMENTS_FONT = "tags"

# named colours in chessboard colour scheme.
LITECOLOR_NAME = "light squares"
DARKCOLOR_NAME = "dark squares"
WHITECOLOR_NAME = "white pieces"
BLACKCOLOR_NAME = "black pieces"
LINE_COLOR_NAME = "line"
MOVE_COLOR_NAME = "move"
ALTERNATIVE_MOVE_COLOR_NAME = "alternative move"
VARIATION_COLOR_NAME = "variation"

# default chessboard colour scheme.
LITECOLOR = "#d9eed9"
DARKCOLOR = "#c2c298"
WHITECOLOR = "#db0ee8"
BLACKCOLOR = "#181634"

# default game score colour scheme.
LINE_COLOR = "#76d9d9"  # a light blue
MOVE_COLOR = "#86d929"  # a light green
ALTERNATIVE_MOVE_COLOR = "#eb3010"  # a dark orange
VARIATION_COLOR = "#e0f113"  # a pale yellow

# names of chess score text management tags.
"""The primary tags are POSITION<suffix> and TOKEN<suffix> where suffix is a
serial number.  Every character is tagged by one TOKEN tag and for each TOKEN
tag there is a corresponding POSITION tag which tags all the non-separator
characters tagged by TOKEN.  There is always at least one trailing separator
character and there may be leading separator characters.  A TOKEN_MARK<suffix>
mark is associated with each TOKEN<suffix> and indicates the insertion point
for editing the token.  An empty token has no position range, and maybe no
POSITION tag, and the TOKEN_MARK is used to initialise the POSITION range.

The NAVIGATE_COMMENT NAVIGATE_TOKEN and NAVIGATE_MOVE tags contain the existing
POSITION ranges for the tokens of the relevant types.

RAV_SEP<suffix> is the only tag which collects TOKEN ranges which are used to
colour variations when there is a choice of moves available (recursive
annotation variations).

Each EDIT_... range indicates which tokens are currently editable.  Note that
only the last move in a line is editable.  The tag name is used to pick the
editing rules applied to the token.

An EDIT_MOVE range must be the last range in a RAV_MOVES tag that is also a
NAVIGATE_MOVE range.

Note that WHITESPACE is not used but is intended for picking out editable
whitespace and will be used, if necessary, as NAVIGATE_WHITESPACE.

RAV... tags refer to all moves between matching PGN RAV tags although only
the relevant characters for the Tk tag's purpose get tagged.
LINE... tags refer to all moves to the right of the current move until the
next unmatched PGN END RAV tag.
VARIATION... tags refer to all moves to the right of the current move until the
next unmatched PGN START RAV tag.
So moves4 and moves6 are in LINE... tags while moves4 and moves6 are in
VARIATION... tags for currentmove in the sequence:
"( moves1 ( moves2 ) moves3 currentmove moves4 ( moves5 ) moves6 )".
This sequence has three sets of RAV... tags: one for the whole sequence and
one each for the sub-sequences "( moves2 )" and "( moves5 )".
"""
# LINE... and VARIATION... got used this way in wxWidgets days so RAV... is
# borrowed from PGN terminology to do the 'whole variation' tasks.  In chess
# cirles the terms line and variation are mostly interchangable, but if talking
# about a game, variations refers to alternatives to the moves actually played.

BUILD_TAG = "bt"  # move and RAV markers (deleted after widget build)
NAVIGATE_COMMENT = "nc"  # can be selected by comment navigation
NAVIGATE_TOKEN = "nt"  # can be selected by navigation
NAVIGATE_MOVE = "nm"  # can be selected by move navigation
EDIT_GLYPH = "eg"  # can be selected for glyph editing
EDIT_RESULT = "er"  # can be selected for result editing
EDIT_PGN_TAG_NAME = "eptn"  # can be selected for PGN Tag name editing
EDIT_PGN_TAG_VALUE = "eptv"  # can be selected for PGN Tag value editing
EDIT_COMMENT = "ec"  # can be selected for comment text editing
EDIT_RESERVED = "erv"  # can be selected for reserved text editing
EDIT_COMMENT_EOL = "eolc"  # can be selected for comment to eol editing
EDIT_ESCAPE_EOL = "eole"  # can be selected for escape to eol editing
EDIT_MOVE_ERROR = "mverr"  # can be selected for move error editing
EDIT_MOVE = "em"  # can be selected for move editing
INSERT_RAV = "ir"  # can be selected for rav insertion
MOVE_EDITED = "me"  # moves that are being edited
WHITESPACE = "ws"  # can be deleted or inserted between editables
RAV_MOVES = "rm"  # prefix for line tags
CHOICE = "ch"  # prefix for line choice tags
SELECTION = "se"  # prefix for line selected choice tags
PRIOR_MOVE = "pm"  # prefix for line prior move tags
RAV_SEP = "rs"  # prefix for line tags with trailing separator
RAV_START_TAG = "rst"  # all ranges for start RAV tokens (added for edit)
RAV_END_TAG = "ret"  # all ranges for end RAV tokens
RAV_TAG = "rt"  # prefix for a variations start and end RAV tags
ALL_CHOICES = "ac"  # all the first moves of RAV_MOVES tags
POSITION = "po"  # prefix for position map tags
TOKEN = "tn"  # prefix for token tags
TOKEN_MARK = "tm"  # prefix for insertion point marks for tokens
PGN_TAG = "pt"  # all ranges for PGN Tag name and value pairs
TERMINATION_TAG = "tt"  # all ranges for values of termination tags
DESCRIPTION = "pd"  # all ranges for name of partial position
PIECE_LOCATIONS = "lo"  # all ranges for piece squares in partial position
# mark the start of the game score, which is also end of PGN tags.
START_SCORE_MARK = "scoremark"
# mark start and end points of editable range, usually range of current.
START_EDIT_MARK = "starteditmark"
END_EDIT_MARK = "endeditmark"
# move played to reach position displayed on board.
MOVE_TAG = MOVE_COLOR_NAME
# moves in selected variation to be played from position displayed on board.
LINE_TAG = LINE_COLOR_NAME
# first move in alternatives to selected variation not yet entered.
ALTERNATIVE_MOVE_TAG = ALTERNATIVE_MOVE_COLOR_NAME
# moves played from main line to reach position in which MOVE_TAG move played.
VARIATION_TAG = VARIATION_COLOR_NAME
# last character in line. Hide the LINE_TAG colour to indicate no more moves.
LINE_END_TAG = "line_end"
# mark the start of the partial position, which is also end of description.
START_POSITION_MARK = "positionmark"
# indentation tag for variations generated by Chess Engines in analysis widget
# of a Game instance.
ANALYSIS_INDENT_TAG = "indent"
# elide tag for PGN tags used in analysis score text widgets.
ANALYSIS_PGN_TAGS_TAG = "analysispgntags"
# indentation tag for wrapped game movetext.
MOVETEXT_INDENT_TAG = "movetextindent"
# elide tag for move numbers in game movetext.
MOVETEXT_MOVENUMBER_TAG = "movenumbertag"
# indentation tag for first move after forced newline in game movetext.
FORCED_INDENT_TAG = "forcedindent"
# newline characters inserted in game movetext for layout or performance.
FORCED_NEWLINE_TAG = "forcednewlinetag"

EMPTY_SEVEN_TAG_ROSTER = "".join(
    [r.join(("[", '""]')) for r in SEVEN_TAG_ROSTER]
)
GRID_HEADER_SEVEN_TAG_ROSTER = frozenset(
    (TAG_WHITE, TAG_BLACK, TAG_RESULT, TAG_EVENT, TAG_DATE)
)
STATUS_SEVEN_TAG_ROSTER_EVENT = (
    TAG_WHITE,
    TAG_RESULT,
    TAG_BLACK,
    TAG_DATE,
    TAG_EVENT,
)
STATUS_SEVEN_TAG_ROSTER_SCORE = (TAG_WHITE, TAG_RESULT, TAG_BLACK, TAG_DATE)
STATUS_SEVEN_TAG_ROSTER_PLAYERS = (TAG_WHITE, TAG_RESULT, TAG_BLACK)
EMPTY_REPERTOIRE_GAME = "".join(
    [r.join(("[", '""]')) for r in REPERTOIRE_TAG_ORDER]
)

SPACE_SEP = " "  # separator used displaying moves and so on
NEWLINE_SEP = "\n"  # separator used displaying moves and so on
NULL_SEP = ""  # separator used displaying moves and so on

# mark the start of the selection rule, which is also end of rule's name.
START_SELECTION_RULE_MARK = "rulemark"

# Force game score text widgets to display every fullmove, a white halfmove and
# a black halfmove, on a new line if the total number of fullmoves including
# variations exceeds this limit.  This is for performance reasons that become
# noticeable at about 500 fullmoves.  Repertoire displays may often exceed this
# limit.  The maximum length game, nearly 6000 fullmoves, takes more than an
# hour to display, or do a navigation operation.
# Force game score text widgets to split movetext into lines, by newline, every
# few moves.  This is for performance reasons that become noticeable at about
# 500 fullmoves.  Repertoire displays may often exceed this limit.  The maximum
# length game, nearly 6000 fullmoves, takes more than an hour to display, or do
# a navigation operation.
# 20 seems reasonable because it does not spread moves out too much and allows
# for move numbers to be inserted without generating too many extra tokens.
FORCE_NEWLINE_AFTER_FULLMOVES = 20



# fullpositionds.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for full position index.

The FullPositionDS class in this module supports the apsw, db, and sqlite3,
interfaces to a database.

See the ..basecore.fullposition module for the FullPositionDS class for apsw,
db, and sqlite3.

"""

from solentware_grid.dpt.datasourcecursor import DataSourceCursor

from ..basecore.fullposition import FullPosition


class FullPositionDS(FullPosition, DataSourceCursor):

    """Combine a DPT DataSourceCursor with FullPosition."""



# analysis.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Interface to chess database for chess engine analysis index.
"""

from ..core.filespec import (
    VARIATION_FIELD_DEF,
    ENGINE_FIELD_DEF,
    ANALYSIS_FILE_DEF,
)


class Analysis:

    """Represent chess engine analysis on file that matches a postion.

    Notes:

    The find_*() methods should migrate to the database engine modules and the
    get_*() methods should migrate to a ../core/? module.

    """

    def __init__(self, dbhome, dbset, dbname, newrow=None):
        super().__init__(dbhome, dbset, dbname, newrow=newrow)

        # FEN and engine used to do analysis.
        self.engine = None
        self.fen = None

    def find_position_analysis(self, fen):
        """Find analysis records matching fen position."""
        self.engine = None
        self.fen = None
        if not fen:
            self.set_recordset(self.dbhome.recordlist_nil(self.dbset))
            return

        fen = self.dbhome.encode_record_selector(fen)

        recordset = self.dbhome.recordlist_key(
            self.dbset, VARIATION_FIELD_DEF, fen
        )

        self.set_recordset(recordset)
        self.fen = fen

    def find_engine_analysis(self, engine):
        """Find analysis records matching chess engine."""
        self.engine = None
        self.fen = None
        if not engine:
            self.set_recordset(self.dbhome.recordlist_nil(self.dbset))
            return

        engine = self.dbhome.encode_record_selector(engine)

        recordset = self.dbhome.recordlist_key(
            self.dbset, ENGINE_FIELD_DEF, engine
        )

        self.set_recordset(recordset)
        self.engine = engine

    def find_engine_position_analysis(self, engine=None, fen=None):
        """Find analysis records matching chess engine and fen."""
        self.engine = None
        self.fen = None
        if not engine:
            if not fen:
                self.set_recordset(self.dbhome.recordlist_nil(self.dbset))
            else:
                self.find_position_analysis(fen)
            return
        elif not fen:
            self.find_engine_analysis(engine)
            return

        engine = self.dbhome.encode_record_selector(engine)
        fen = self.dbhome.encode_record_selector(fen)

        fenset = self.dbhome.recordlist_key(
            self.dbset, VARIATION_FIELD_DEF, fen
        )
        engineset = self.dbhome.recordlist_key(
            self.dbset, ENGINE_FIELD_DEF, engine
        )
        self.set_recordset(engineset & fenset)

        self.engine = engine
        self.fen = fen

    def get_position_analysis(self, fen):
        """Get analysis matching fen position.

        It is assumed merging data from all records matching fen makes sense.

        """
        self.find_position_analysis(fen)
        analysis = self.newrow().value
        ar = self.newrow()
        arv = ar.value
        rsc = self.get_cursor()
        try:
            r = rsc.first()
            while r:
                ar.load_record(r)
                analysis.scale.update(arv.scale)
                analysis.variations.update(arv.variations)
                r = rsc.next()
            else:
                analysis.position = fen
        finally:
            rsc.close()
        return analysis

    def get_position_analysis_records(self, fen):
        """Return list of analysis records matching fen position."""
        self.find_position_analysis(fen)
        records = []
        rsc = self.get_cursor()
        try:
            r = rsc.first()
            while r:
                ar = self.newrow()
                ar.load_record(r)
                records.append(ar)
                r = rsc.next()
        finally:
            rsc.close()
        return records



# uci_to_pgn.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Convert UCI Chess Engine moves to PGN.

UCI chess engines give moves as <from square><to square> and the easiest way of
handling these, given the existing PGN parser, is translate to PGN.

generate_pgn_for_uci_moves_in_position() generates the legal unambiguous PGN
description of a sequence of UCI Chess Engine move for the position it refers
to.

'Qd1f3' is always unambiguous but is the legal unambiguous description only if
more than two queens of the side with the move can legally move to 'f3'. 'Qf3'
is usually the only legal PGN description because the starting position for a
game has one queen per side. 'Qf3' is not necessarily a legal move.

"""
import re

from pgn_read.core.constants import (
    FEN_WHITE_KING,
    FEN_WHITE_QUEEN,
    FEN_WHITE_ROOK,
    FEN_WHITE_BISHOP,
    FEN_WHITE_KNIGHT,
    FEN_WHITE_PAWN,
    FEN_BLACK_KING,
    FEN_BLACK_QUEEN,
    FEN_BLACK_ROOK,
    FEN_BLACK_BISHOP,
    FEN_BLACK_KNIGHT,
    FEN_BLACK_PAWN,
    PGN_KING,
    PGN_QUEEN,
    PGN_ROOK,
    PGN_BISHOP,
    PGN_KNIGHT,
    PGN_PAWN,
    OTHER_SIDE,
    FEN_WHITE_PIECES,
    FEN_BLACK_PIECES,
    FEN_WHITE_ACTIVE,
    FEN_BLACK_ACTIVE,
    PGN_CAPTURE_MOVE,
)
from pgn_read.core.parser import add_token_to_game

from .pgn import GameMove
from .constants import (
    NOPIECE,
    FEN_CONTEXT,
)

_PIECE_TO_PGN = {
    FEN_WHITE_KING: PGN_KING,
    FEN_WHITE_QUEEN: PGN_QUEEN,
    FEN_WHITE_ROOK: PGN_ROOK,
    FEN_WHITE_BISHOP: PGN_BISHOP,
    FEN_WHITE_KNIGHT: PGN_KNIGHT,
    FEN_WHITE_PAWN: PGN_PAWN,
    FEN_BLACK_KING: PGN_KING,
    FEN_BLACK_QUEEN: PGN_QUEEN,
    FEN_BLACK_ROOK: PGN_ROOK,
    FEN_BLACK_BISHOP: PGN_BISHOP,
    FEN_BLACK_KNIGHT: PGN_KNIGHT,
    FEN_BLACK_PAWN: PGN_PAWN,
}
_PROMOTE = {
    PGN_QUEEN.lower(): "".join(("=", PGN_QUEEN)),
    PGN_ROOK.lower(): "".join(("=", PGN_ROOK)),
    PGN_BISHOP.lower(): "".join(("=", PGN_BISHOP)),
    PGN_KNIGHT.lower(): "".join(("=", PGN_KNIGHT)),
    "": "",
}
_CASTLES = {"e1g1": "O-O", "e8g8": "O-O", "e1c1": "O-O-O", "e8c8": "O-O-O"}
_CASTLEKEY = {
    "e1g1": FEN_WHITE_KING,
    "e8g8": FEN_BLACK_KING,
    "e1c1": FEN_WHITE_KING,
    "e8c8": FEN_BLACK_KING,
}
_ACTIVE_PIECES = {
    FEN_WHITE_ACTIVE: FEN_WHITE_PIECES,
    FEN_BLACK_ACTIVE: FEN_BLACK_PIECES,
}

re_move = re.compile(
    "".join(("^", "([a-h][1-8])", "([a-h][1-8])", "([qrbn]?)", "$"))
)


def generate_pgn_for_uci_moves_in_position(moves, fen):
    """Return PGN-style movetext and update position for unambiguous moves."""
    game = GameMove()
    piece_placement_data = game._piece_placement_data
    text = []
    try:
        moves = moves.split()
    except:
        return "".join(
            ("{'", str(moves), "' cannot be a move, 'Yz0' inserted.}Yz0")
        )
    if not moves:
        return "{'' is not a move, 'Yz0' inserted. Rest '' ignored.}Yz0"
    tagtext = fen.join(FEN_CONTEXT)
    match_end = add_token_to_game(tagtext, game)
    match_end = add_token_to_game(tagtext, game, pos=match_end)
    if not game.set_initial_position():
        return "".join(
            (
                "{'Forsyth-Edwards Notation sets an illegal position. ",
                "Move 'Yz0' inserted.}Yz0",
            )
        )
    for count, move in enumerate(moves):
        g = re_move.match(move)
        if not g:
            text.append(
                "".join(
                    (
                        "{'",
                        str(move),
                        "' cannot be a move, 'Yz0' inserted. Rest '",
                        " ".join(moves[count + 1 :]),
                        "' ignored.}Yz0",
                    )
                )
            )
            break
        from_square, to_square, promote_to = g.groups()

        # from_square must contain a piece belonging to side with the move.
        if from_square not in piece_placement_data:
            text.append(
                "".join(
                    (
                        "{'",
                        str(move),
                        "' does not refer to a piece of the active side, ",
                        "'Yz0' inserted. Rest '",
                        " ".join(moves[count + 1 :]),
                        "' ignored.}Yz0",
                    )
                )
            )
            break
        piece = piece_placement_data[from_square].name
        if piece not in _ACTIVE_PIECES[game._active_color]:
            text.append(
                "".join(
                    (
                        "{'",
                        str(move),
                        "' does not refer to a piece of the active side, ",
                        "'Yz0' inserted. Rest '",
                        " ".join(moves[count + 1 :]),
                        "' ignored.}Yz0",
                    )
                )
            )
            break

        # to_square must not contain a piece belonging to side with the move.
        if to_square in piece_placement_data:
            destination_piece = piece_placement_data[to_square].name
            if destination_piece in _ACTIVE_PIECES[game._active_color]:
                text.append(
                    "".join(
                        (
                            "{'",
                            str(move),
                            "' cannot be a move, 'Yz0' inserted. Rest '",
                            " ".join(moves[count + 1 :]),
                            "' ignored.}Yz0",
                        )
                    )
                )
                break

        # Maybe the _PIECE_TO_PGN mapping should be replaced by an attribute
        # in the Piece class.
        piece = _PIECE_TO_PGN[piece]

        promote_to = _PROMOTE[promote_to]

        # What about illegal moves specified in e2f5 format where the
        # constructed PGN move happens to be legal.

        # Castles is treated as a king move of two squares, otherwise illegal.
        if move in _CASTLES and piece_placement_data[
            from_square
        ].name == _CASTLEKEY.get(move):
            pgn_movetext = _CASTLES[move]

        # Pawn moves are given in a mixture of short algebraic notation and
        # long algebraic notation to fit the expectations of pgn_read's Game
        # class (at time of writing).  The move text returned by the chess
        # engine, <source square><destination square>, is used directly as
        # long algebraic notation passed to the PGN parser because it fits
        # the processing of, for example, 'g7g8' when a black pawn is on 'g7'.
        # Promotions and captures fit processing as short algebraic notation.
        # Short algebraic notation is generated by the GameMove class.
        elif piece == PGN_PAWN:

            # The distinction between normal captures, to_square occupied, and
            # en-passant captures, to_square empty, is left to the PGN parser.
            if from_square[0] == to_square[0]:
                if promote_to:
                    pgn_movetext = to_square + promote_to
                else:
                    pgn_movetext = from_square + to_square
            else:
                pgn_movetext = "".join(
                    (from_square[0], PGN_CAPTURE_MOVE, to_square, promote_to)
                )

            # Pawn must not be promoted to ranks othen than 1 or 8.
            if promote_to:
                if to_square[1] not in "18":
                    text.append(
                        "".join(
                            (
                                "{'",
                                str(move),
                                "' cannot be a move, 'Yz0' inserted. Rest '",
                                " ".join(moves[count + 1 :]),
                                "' ignored.}Yz0",
                            )
                        )
                    )
                    break

        # Piece moves are given in long algebraic notation, relying on the
        # GameMove class to produce short algebraic notation if possible.
        elif to_square not in piece_placement_data:
            pgn_movetext = "".join((piece, from_square, to_square))
        else:
            pgn_movetext = "".join(
                (piece, from_square, PGN_CAPTURE_MOVE, to_square)
            )

        # Generally the 'if matchend ...' should be 'while matchend ...', but
        # pgn_movetext is a single move so two calls to add_token_to_game are
        # needed, at most, to handle moves expressed as 'Qe3c6' for example.
        # This function does not have to figure which of 'Qe3c6', 'Q3c6',
        # 'Qec6', and 'Qc6', is the appropriate PGN movetext in the position.
        matchend = add_token_to_game(pgn_movetext, game)
        if matchend != len(pgn_movetext):
            add_token_to_game(pgn_movetext, game, pos=matchend)

        if game.state is not None:
            text.append(
                "".join(
                    (
                        "{'",
                        str(move),
                        "' cannot be a move, 'Yz0' inserted. Rest '",
                        " ".join(moves[count + 1 :]),
                        "' ignored.}Yz0",
                    )
                )
            )
            break
        text.append(game._text[-1])
    return " ".join(text)
